# Standard
from typing import Tuple
import contextlib
import logging
import multiprocessing
import pathlib
import socket
import typing

# Local
from ...configuration import get_model_family

logger = logging.getLogger(__name__)

API_ROOT_WELCOME_MESSAGE = "Hello from InstructLab! Visit us at https://instructlab.ai"
CHAT_TEMPLATE_AUTO = "auto"
CHAT_TEMPLATE_TOKENIZER = "tokenizer"
LLAMA_CPP = "llama_cpp"
VLLM = "vllm"
templates = [
    {
        "model": "merlinite",
        "template": "{% for message in messages %}\n{% if message['role'] == 'user' %}\n{{ '<|user|>\n' + message['content'] }}\n{% elif message['role'] == 'system' %}\n{{ '<|system|>\n' + message['content'] }}\n{% elif message['role'] == 'assistant' %}\n{{ '<|assistant|>\n' + message['content'] + eos_token }}\n{% endif %}\n{% if loop.last and add_generation_prompt %}\n{{ '<|assistant|>' }}\n{% endif %}\n{% endfor %}",
    },
    {
        "model": "mixtral",
        "template": "{{ bos_token }}\n{% for message in messages %}\n{% if message['role'] == 'user' %}\n{{ '[INST] ' + message['content'] + ' [/INST]' }}\n{% elif message['role'] == 'assistant' %}\n{{ message['content'] + eos_token}}\n{% endif %}\n{% endfor %}",
    },
]


class Closeable(typing.Protocol):
    def close(self) -> None: ...


class ServerException(Exception):
    """An exception raised when serving the API."""


def get_model_template(
    model_family: str, model_path: pathlib.Path
) -> Tuple[str, str, str]:
    eos_token = "<|endoftext|>"
    bos_token = ""
    template = ""
    resolved_family = get_model_family(model_family, model_path)
    logger.debug(
        "Searching hard coded model templates for model family %s's template",
        resolved_family,
    )
    for template_dict in templates:
        if template_dict["model"] == resolved_family:
            template = template_dict["template"]
            if template_dict["model"] == "mixtral":
                eos_token = "</s>"
                bos_token = "<s>"

    return template, eos_token, bos_token


def is_temp_server_running():
    """Check if the temp server is running."""
    return multiprocessing.current_process().name != "MainProcess"


def verify_template_exists(path):
    if not path.exists():
        raise FileNotFoundError("Chat template file does not exist: {}".format(path))

    if not path.is_file():
        raise IsADirectoryError(
            "Chat templates paths must point to a file: {}".format(path)
        )


def free_tcp_ipv4_port(host: str) -> int:
    """Ask the OS for a random, ephemeral, and bindable TCP/IPv4 port

    Note: The idea of finding a free port is bad design and subject to
    race conditions. Instead vLLM and llama-cpp should accept port 0 and
    have an API to return the actual listening port. Or they should be able
    to use an existing socket like a systemd socket activation service.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((host, 0))
        return int(s.getsockname()[-1])


def safe_close_all(resources: typing.Iterable[Closeable]):
    for resource in resources:
        with contextlib.suppress(Exception):
            resource.close()



# SPDX-License-Identifier: Apache-2.0
# pylint: disable=duplicate-code

# Standard
from unittest.mock import MagicMock, patch
import os

# Third Party
from click.testing import CliRunner
from git import GitError

# First Party
from instructlab import lab
from instructlab.configuration import DEFAULTS, read_config


class TestLabInit:
    @patch(
        "instructlab.config.init.convert_bytes_to_proper_mag", return_value=(89.0, "GB")
    )
    def test_ilab_config_init_auto_detection(
        self, convert_bytes_to_proper_mag_mock, cli_runner: CliRunner
    ):
        result = cli_runner.invoke(lab.ilab, ["config", "init", "--interactive"])
        assert result.exit_code == 0, result.stdout
        convert_bytes_to_proper_mag_mock.assert_called_once()
        assert (
            "We chose Nvidia 2x A100 as your designated training profile. This is for systems with 160 GB of vRAM"
            in result.stdout
        )

    # When using `from X import Y` you need to understand that Y becomes part
    # of your module, so you should use `my_module.Y`` to patch.
    # When using `import X`, you should use `X.Y` to patch.
    # https://docs.python.org/3/library/unittest.mock.html#where-to-patch?
    @patch("instructlab.config.init.Repo.clone_from")
    def test_init_noninteractive(self, mock_clone_from, cli_runner: CliRunner):
        result = cli_runner.invoke(lab.ilab, ["config", "init", "--non-interactive"])
        assert result.exit_code == 0
        assert os.path.exists(DEFAULTS.CONFIG_FILE)
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        mock_clone_from.assert_called_once()

    @patch("instructlab.config.init.Repo.clone_from")
    def test_init_interactive(self, mock_clone_from, cli_runner: CliRunner):
        result = cli_runner.invoke(lab.ilab, ["config", "init"], input="\nn")
        assert result.exit_code == 0
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        assert "taxonomy" not in os.listdir()
        mock_clone_from.assert_not_called()

    @patch(
        "instructlab.config.init.Repo.clone_from",
        MagicMock(side_effect=GitError("Authentication failed")),
    )
    def test_init_interactive_git_error(self, cli_runner: CliRunner):
        result = cli_runner.invoke(lab.ilab, ["config", "init"], input="y\n\ny")
        assert result.exit_code == 1, "command finished with an unexpected exit code"
        assert "Failed to clone taxonomy repo: Authentication failed" in result.output
        assert "manually run" in result.output

    @patch("instructlab.config.init.Repo.clone_from")
    def test_init_interactive_clone(self, mock_clone_from, cli_runner: CliRunner):
        result = cli_runner.invoke(lab.ilab, ["config", "init"], input="y\n\ny")
        assert result.exit_code == 0
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        mock_clone_from.assert_called_once()

    @patch("instructlab.config.init.Repo.clone_from")
    def test_init_interactive_default_clone(
        self, mock_clone_from, cli_runner: CliRunner
    ):
        result = cli_runner.invoke(lab.ilab, ["config", "init"], input="\n")
        assert result.exit_code == 0
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        mock_clone_from.assert_called_once()

    @patch("instructlab.config.init.Repo.clone_from")
    def test_init_interactive_with_preexisting_nonempty_taxonomy(
        self, mock_clone_from, cli_runner: CliRunner
    ):
        os.makedirs(f"{DEFAULTS.TAXONOMY_DIR}/contents")
        result = cli_runner.invoke(lab.ilab, ["config", "init"], input="\n\n")
        assert result.exit_code == 0
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        assert "taxonomy" in os.listdir(DEFAULTS._data_dir)
        mock_clone_from.assert_not_called()

    def test_init_interactive_with_preexisting_config(self, cli_runner: CliRunner):
        result = cli_runner.invoke(
            lab.ilab, ["config", "init"], input="non-default-taxonomy\nn"
        )
        assert result.exit_code == 0
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        config = read_config(DEFAULTS.CONFIG_FILE)
        assert config.generate.taxonomy_path == "non-default-taxonomy"

        # second invocation should ask if we want to overwrite - yes, and change taxonomy path
        result = cli_runner.invoke(
            lab.ilab, ["config", "init"], input="y\ny\ndifferent-taxonomy\nn"
        )
        assert result.exit_code == 0
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        config = read_config(DEFAULTS.CONFIG_FILE)
        assert config.generate.taxonomy_path == "different-taxonomy"

        # third invocation should again ask, but this time don't overwrite
        result = cli_runner.invoke(lab.ilab, ["config", "init"], input="n")
        assert result.exit_code == 0
        assert "config.yaml" in os.listdir(DEFAULTS._config_dir)
        config = read_config(DEFAULTS.CONFIG_FILE)
        assert config.generate.taxonomy_path == "different-taxonomy"



"""Helper for test_ilab_cli_imports"""

# Standard
from importlib import metadata
import sys

# block slow imports
for unwanted in ["deepspeed", "llama_cpp", "torch", "vllm"]:
    # importlib raises ModuleNotFound when sys.modules value is None.
    assert unwanted not in sys.modules
    sys.modules[unwanted] = None  # type: ignore[assignment]

for command in metadata.entry_points(group="instructlab.command"):
    print(f"{command.name}")
    command.load()
    for subcommand in metadata.entry_points(
        group=f"instructlab.command.{command.name}"
    ):
        print(f"  {command.name}.{subcommand.name}")
        subcommand.load()



# SPDX-License-Identifier: Apache-2.0
# Standard
import sys

# Third Party
from ruamel.yaml import YAMLError, ruamel
import click

# First Party
from instructlab import clickext, configuration

# Initialize ruamel.yaml
yaml = ruamel.yaml.YAML()
yaml.indent(mapping=2, sequence=4, offset=2)


@click.command()
@click.pass_context
@clickext.display_params
def show(ctx: click.Context) -> None:
    """Displays the current config as YAML"""
    # TODO: make this use pretty colors like jq/yq
    try:
        commented_map = configuration.config_to_commented_map(ctx.obj.config)
    except YAMLError as e:
        click.secho(f"Error loading config as YAML: {e}", fg="red")
        ctx.exit(2)
    yaml.dump(commented_map, sys.stdout)



"""Do not copy SELinux context

Workaround for https://github.com/python/cpython/issues/83074
"""

# sitecustomize is imported by site.py on interpreter startup
# Standard
import errno
import os
import shutil

if hasattr(os, "listxattr"):

    def _patched_copyxattr(src, dst, *, follow_symlinks=True):
        """Copy extended filesystem attributes from `src` to `dst`.

        Overwrite existing attributes.

        If `follow_symlinks` is false, symlinks won't be followed.
        """
        try:
            names = os.listxattr(src, follow_symlinks=follow_symlinks)
        except OSError as e:
            if e.errno not in (errno.ENOTSUP, errno.ENODATA, errno.EINVAL):
                raise
            return

        # WORKAROUND: do not copy system, security, or trusted attributes.
        # Only copy user attributes.
        names = [name for name in names if name.startswith("user.")]

        for name in names:
            try:
                value = os.getxattr(src, name, follow_symlinks=follow_symlinks)
                os.setxattr(dst, name, value, follow_symlinks=follow_symlinks)
            except OSError as e:
                if e.errno not in (
                    errno.EPERM,
                    errno.ENOTSUP,
                    errno.ENODATA,
                    errno.EINVAL,
                    errno.EACCES,
                ):
                    raise

    shutil._copyxattr = _patched_copyxattr



# SPDX-License-Identifier: Apache-2.0

# Standard
from contextlib import redirect_stderr
from time import sleep
from types import FrameType
from typing import Optional, cast
import logging
import multiprocessing
import pathlib
import signal

# Third Party
from llama_cpp import llama_chat_format, llama_token_get_text
from llama_cpp.server.app import create_app
from llama_cpp.server.model import LlamaProxy
from llama_cpp.server.settings import Settings
from uvicorn import Config
import fastapi
import httpx
import llama_cpp.server.app as llama_app
import uvicorn

# Local
from ...client import check_api_base
from ...configuration import get_api_base
from .common import (
    API_ROOT_WELCOME_MESSAGE,
    CHAT_TEMPLATE_AUTO,
    CHAT_TEMPLATE_TOKENIZER,
    LLAMA_CPP,
    ServerException,
    free_tcp_ipv4_port,
    get_model_template,
    is_temp_server_running,
    verify_template_exists,
)
from .server import BackendServer, ServerConfig

logger = logging.getLogger(__name__)


class Server(BackendServer):
    def __init__(
        self,
        model_path: pathlib.Path,
        model_family: str,
        chat_template: str,
        api_base: str,
        host: str,
        port: int,
        gpu_layers: int,
        max_ctx_size: int,
        num_threads: Optional[int],
        log_file: Optional[pathlib.Path] = None,
    ):
        sc = ServerConfig(api_base, log_file)
        super().__init__(
            model_family,
            model_path,
            chat_template,
            host,
            port,
            sc,
        )
        self.gpu_layers = gpu_layers
        self.max_ctx_size = max_ctx_size
        self.num_threads = num_threads
        self.queue: Optional[multiprocessing.Queue] = None
        self.process: multiprocessing.Process | None = None

    def run(self):
        """Start an OpenAI-compatible server with llama-cpp"""
        try:
            server(
                model_path=self.model_path,
                chat_template=self.chat_template,
                gpu_layers=self.gpu_layers,
                max_ctx_size=self.max_ctx_size,
                model_family=self.model_family,
                threads=self.num_threads,
                host=self.host,
                port=self.port,
                log_file=self.config.log_file,
                log_level=logger.getEffectiveLevel(),
            )
        except ServerException as exc:
            raise exc

    def create_server_process(self, port: int) -> multiprocessing.Process:
        mpctx = multiprocessing.get_context(None)
        self.queue = mpctx.Queue()

        # When using the multiprocessing module, each new process starts with a fresh copy of
        # the parent process's state, but it does not inherit the state of the logging configuration.
        # This means that any logging configuration set up in the parent process (such as log levels,
        # handlers, etc.) will not be automatically applied to the child processes.
        # That's why we pass the log_level parameter to the server function and set the log level
        # We only do this when called from run_detached
        server_process = mpctx.Process(
            target=server,
            kwargs={
                "model_path": self.model_path,
                "chat_template": self.chat_template,
                "gpu_layers": self.gpu_layers,
                "max_ctx_size": self.max_ctx_size,
                "model_family": self.model_family,
                "port": port,
                "host": self.host,
                "queue": self.queue,
                "log_file": self.config.log_file,
                "log_level": logger.getEffectiveLevel(),
            },
        )

        return server_process

    def run_detached(
        self,
        http_client: httpx.Client | None = None,
        background: bool = True,
        foreground_allowed: bool = False,
        max_startup_retries: int = 0,
    ) -> str:
        logger.info(f"Trying to connect to model server at {self.config.api_base}")
        if check_api_base(self.config.api_base, http_client):
            return self.config.api_base
        try:
            self.port = free_tcp_ipv4_port(self.host)
            # start new server
            self.api_base = str(get_api_base(f"{self.host}:{self.port}"))
            logger.debug(f"Starting a temporary server at {self.api_base}...")
            self.process = self.create_server_process(self.port)
            self.process.start()

            # in case the server takes some time to fail we wait a bit
            logger.debug("Waiting for the server to start...")
            count = 0
            while self.process.is_alive():
                sleep(0.1)
                if check_api_base(self.api_base, http_client):
                    break
                if count > 50:
                    logger.error("failed to reach the API server")
                    break
                count += 1

            logger.debug("Server started.")

            # if the queue is not empty it means the server failed to start
            if self.queue is not None and not self.queue.empty():
                # pylint: disable=raise-missing-from
                raise self.queue.get()

        except ServerException as exc:
            self.shutdown()
            raise exc
        return self.api_base

    def shutdown(self):
        """Stop the server process and close the queue."""

        super().shutdown()

        if self.process and self.queue:
            self.process.terminate()
            self.process.join(timeout=30)
            self.queue.close()
            self.queue.join_thread()

    def get_backend_type(self):
        return LLAMA_CPP


def server(
    model_path: pathlib.Path,
    chat_template: str,
    gpu_layers: int,
    max_ctx_size: int,
    model_family: str,
    threads=None,
    host: str = "localhost",
    port: int = 8000,
    queue: Optional[multiprocessing.Queue] = None,
    log_file: Optional[pathlib.Path] = None,
    log_level: int = logging.INFO,
):
    """Start OpenAI-compatible server"""
    verbose = log_level == logging.DEBUG
    settings = Settings(
        host=host,
        port=port,
        model=model_path.as_posix(),
        n_ctx=max_ctx_size,
        n_gpu_layers=gpu_layers,
        verbose=verbose,
    )

    if threads is not None:
        settings.n_threads = threads
    try:
        # When we run a logger with DEBUG, verbose mode is activated, create_app will initialize the Llama class which
        # will print the model configuration to stderr. We need to redirect stderr to the log_file
        # if specified.
        # We don't need to redirect if verbose is not true, since there will be nothing to redirect.
        if verbose and log_file:
            # We can't redirect to the logger here because we might end up in a RecursionError. It
            # is likely caused by the logger itself writing to stderr, which is being redirected to
            # the logger, creating a recursive loop.
            with (
                open(log_file, "a", encoding="utf-8") as f,
                redirect_stderr(f),
            ):
                app = create_app(settings=settings)
        else:
            app = create_app(settings=settings)

        @app.get("/")
        def read_root():
            return {"message": API_ROOT_WELCOME_MESSAGE}
    except ValueError as exc:
        if queue:
            queue.put(exc)
            queue.close()
            queue.join_thread()
            return
        raise ServerException(f"failed creating the server application: {exc}") from exc

    # Update chat template if necessary
    augment_chat_template(chat_template, model_family, model_path, queue)

    logger.info("Starting server process, press CTRL+C to shutdown server...")
    logger.info(
        f"After application startup complete see http://{host}:{port}/docs for API."
    )

    # Build server's configuration
    config = get_uvicorn_config(
        app=app,
        host=host,
        port=port,
    )
    s = UvicornServer(config)

    # If this is not the main process, this is the temp server process that ran in the background
    # after `ilab model chat` was executed.
    # In this case, we want to redirect stdout to null to avoid cluttering the chat with messages
    # returned by the server.
    # Redirect stderr to log file if specified
    if log_file:
        with open(log_file, "a", encoding="utf-8") as f, redirect_stderr(f):
            # Start the server
            s.run()
    else:
        s.run()

    if queue:
        queue.close()
        queue.join_thread()


def get_uvicorn_config(app: fastapi.FastAPI, host: str, port: int) -> Config:
    return Config(
        app,
        host=host,
        port=port,
        log_level=logging.ERROR,
        limit_concurrency=2,  # Make sure we only serve a single client at a time
        timeout_keep_alive=0,  # prevent clients holding connections open (we only have 1)
    )


class UvicornServer(uvicorn.Server):
    """Override uvicorn.Server to handle SIGINT."""

    def handle_exit(self, sig: int, frame: Optional[FrameType]) -> None:
        if not is_temp_server_running() or sig != signal.SIGINT:
            super().handle_exit(sig=sig, frame=frame)


def augment_chat_template(
    chat_template: str,
    model_family: str,
    model_path: pathlib.Path,
    queue: Optional[multiprocessing.Queue],
):
    # chat template takes the format ('auto' | 'tokenizer' | a filesystem path to a file)
    if chat_template == CHAT_TEMPLATE_TOKENIZER:
        # llama_cpp_python calculates from the tokenizer config on load, so
        # nothing to do
        return

    try:
        if chat_template == CHAT_TEMPLATE_AUTO:
            # Currently "auto" maps to replacing with ilab stored templates
            template, eos_token, bos_token = get_model_template(
                model_family, model_path
            )
        else:
            # In this case, the template is a path to a file; attempt to load it
            template = load_template(chat_template)
            eos_token = None
            bos_token = None

        logger.info("Replacing chat template:\n %s", template)

        for proxy in llama_app.get_llama_proxy():
            proxy().chat_handler = llama_chat_format.Jinja2ChatFormatter(
                template=template,
                # Use the model defined eos and bos if either is not
                # defined (when a custom template is used)
                eos_token=resolve_token_eos(eos_token, proxy),
                bos_token=resolve_token_bos(bos_token, proxy),
            ).to_chat_handler()
    # pylint: disable=broad-exception-caught
    except Exception as exc:
        if queue:
            queue.put(exc)
            queue.close()
            queue.join_thread()
            return
        raise ServerException(f"failed creating the server application: {exc}") from exc


def resolve_token_eos(eos_token: Optional[str], proxy: LlamaProxy) -> str:
    if eos_token is not None:
        return eos_token

    return resolve_token(proxy, proxy().token_eos())


def resolve_token_bos(bos_token: Optional[str], proxy: LlamaProxy) -> str:
    if bos_token is not None:
        return bos_token

    return resolve_token(proxy, proxy().token_bos())


def resolve_token(proxy: LlamaProxy, token: int) -> str:
    result = llama_token_get_text(proxy().model, token).decode("utf-8")
    return cast(str, result)


def load_template(template_path: str) -> str:
    path = pathlib.Path(template_path)
    verify_template_exists(path)

    try:
        return path.read_text(encoding="utf-8")
    except Exception as e:
        raise IOError("Error reading chat template file contents: {}".format(e)) from e



# Standard
from unittest import mock
from unittest.mock import patch
import os
import textwrap

# Third Party
from click.testing import CliRunner
from git import Repo

# First Party
from instructlab import lab

# Local
from . import common


def gen_qa_pairs(odd):
    i = 1
    qa_pairs = []
    score = 0
    while i < 5:
        if i % 2:
            if odd:
                score = 0.2
            else:
                score = 0.1
        elif not i % 2:
            if odd:
                score = 0.3
            else:
                score = 0.4
        qa_pairs.append(
            {
                "question_id": i,
                "score": score,
                "qna_file": f"category{i}/qna.yaml",
            }
        )
        i = i + 1
    qa_pairs.append(
        {
            "question_id": i,
            "score": 0.5,
            "qna_file": f"category{i}/qna.yaml",
        }
    )
    if odd:
        qa_pairs.append(
            {
                "question_id": i + 1,
                "score": 0.6,
                "qna_file": f"category{i+1}/qna.yaml",
            }
        )
    return qa_pairs


def gen_individual_scores(odd):
    i = 1
    individual_scores = {}
    score = 0
    while i < 5:
        if i % 2:
            if odd:
                score = 0.2
            else:
                score = 0.1
        elif not i % 2:
            if odd:
                score = 0.3
            else:
                score = 0.4
        individual_scores[f"task{i}"] = {"score": score}
        i = i + 1
    individual_scores[f"task{i}"] = {"score": 0.5}
    return individual_scores


def setup_taxonomy(taxonomy_path):
    repo = Repo.init(taxonomy_path, initial_branch="main")
    file_name = "README.md"
    file_path = os.path.join(taxonomy_path, file_name)
    with open(file_path, "w", encoding="utf-8") as f:
        f.write("#Taxonomy")
    repo.index.add([file_name])
    repo.index.commit("initial commit")


def run_mt_bench(cli_runner, error_rate):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench",
            "--model",
            "instructlab/granite-7b-lab",
            "--judge-model",
            "instructlab/merlinite-7b-lab",
        ],
    )
    if result.exit_code != 0:
        print(result.output)
    assert result.exit_code == 0
    expected = textwrap.dedent(
        """\
        Generating answers...
        Evaluating answers...
        # SKILL EVALUATION REPORT

        ## MODEL (SCORE)
        instructlab/granite-7b-lab (1.5/10.0)

        ### TURN ONE (0.0 to 10.0):
        1.0

        ### TURN TWO (0.0 to 10.0):
        2
        """
    )
    if error_rate > 0:
        expected += textwrap.dedent(
            f"""\

            ### ERROR RATE:
            {round(error_rate, 2)}
            """
        )
    assert result.output == expected


def run_mt_bench_branch(cli_runner, error_rate):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench_branch",
            "--model",
            "instructlab/granite-7b-lab",
            "--judge-model",
            "instructlab/merlinite-7b-lab",
            "--base-model",
            "instructlab/granite-7b-lab",
            "--branch",
            "rc",
            "--base-branch",
            "main",
            "--taxonomy-path",
            "taxonomy",
        ],
    )
    if result.exit_code != 0:
        print(result.output)
    assert result.exit_code == 0
    expected = textwrap.dedent(
        """\
        Generating questions and reference answers from qna files for branch rc...
        Generating questions and reference answers from qna files for branch main...
        Evaluating answers for branch rc...
        Evaluating answers for branch main...
        # SKILL EVALUATION REPORT

        ## BASE MODEL (SCORE)
        instructlab/granite-7b-lab (0/10.0)

        ## MODEL (SCORE)
        instructlab/granite-7b-lab (0/10.0)

        ### IMPROVEMENTS (0.0 to 10.0):
        1. category1/qna.yaml: 0.1 -> 0.2 (+0.1)
        2. category3/qna.yaml: 0.1 -> 0.2 (+0.1)

        ### REGRESSIONS (0.0 to 10.0):
        1. category2/qna.yaml: 0.4 -> 0.3 (-0.1)
        2. category4/qna.yaml: 0.4 -> 0.3 (-0.1)

        ### NO CHANGE (0.0 to 10.0):
        1. category5/qna.yaml (0.5)

        ### NEW (0.0 to 10.0):
        1. category6/qna.yaml (0.6)
        """
    )
    if error_rate > 0:
        expected += textwrap.dedent(
            f"""\

            ### ERROR RATE:
            {round(error_rate, 2)}
            """
        )
    assert result.output == expected


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.model.evaluate.launch_server",
    return_value=(mock.MagicMock(), "http://127.0.0.1:8000/v1", 1),
)
@patch("instructlab.eval.mt_bench.MTBenchEvaluator.gen_answers")
@patch(
    "instructlab.eval.mt_bench.MTBenchEvaluator.judge_answers",
    side_effect=(
        (1.5001, [{}, {}], [1.002, 2], 0),
        (1.5001, [{}, {}], [1.002, 2], 0.768),
    ),
)
def test_evaluate_mt_bench(
    judge_answers_mock,
    gen_answers_mock,
    launch_server_mock,
    validate_model_mock,
    cli_runner: CliRunner,
):
    run_mt_bench(cli_runner, 0)
    assert validate_model_mock.call_count == 2
    judge_answers_mock.assert_called_once()
    gen_answers_mock.assert_called_once()
    assert launch_server_mock.call_count == 2
    run_mt_bench(cli_runner, 0.768)
    assert validate_model_mock.call_count == 4
    assert judge_answers_mock.call_count == 2
    assert gen_answers_mock.call_count == 2
    assert launch_server_mock.call_count == 4


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.model.evaluate.launch_server",
    return_value=(mock.MagicMock(), "http://127.0.0.1:8000/v1", 1),
)
@patch("instructlab.eval.mt_bench.MTBenchBranchEvaluator.gen_answers")
@patch(
    "instructlab.eval.mt_bench.MTBenchBranchEvaluator.judge_answers",
    side_effect=[
        (0, gen_qa_pairs(True), 0),
        (0, gen_qa_pairs(False), 0),
        (0, gen_qa_pairs(True), 0.4567),
        (0, gen_qa_pairs(False), 0.4567),
    ],
)
def test_evaluate_mt_bench_branch(
    judge_answers_mock,
    gen_answers_mock,
    launch_server_mock,
    validate_model_mock,
    cli_runner: CliRunner,
):
    run_mt_bench_branch(cli_runner, 0)
    assert validate_model_mock.call_count == 3
    assert judge_answers_mock.call_count == 2
    assert gen_answers_mock.call_count == 2
    assert launch_server_mock.call_count == 3
    run_mt_bench_branch(cli_runner, 0.4567)
    assert validate_model_mock.call_count == 6
    assert judge_answers_mock.call_count == 4
    assert gen_answers_mock.call_count == 4
    assert launch_server_mock.call_count == 6


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.model.evaluate.launch_server",
    return_value=(mock.MagicMock(), "http://127.0.0.1:8000/v1", 1),
)
@patch(
    "instructlab.eval.mmlu.MMLUEvaluator.run",
    return_value=(0.5, {"task1": {"score": 0.1}, "task2": {"score": 0.9}}),
)
def test_evaluate_mmlu(
    run_mock, launch_server_mock, validate_model_mock, cli_runner: CliRunner
):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mmlu",
            "--model",
            "instructlab/granite-7b-lab",
        ],
    )
    validate_model_mock.assert_called_once()
    launch_server_mock.assert_called_once()
    run_mock.assert_called_once()
    assert result.exit_code == 0
    expected = textwrap.dedent(
        """\
        # KNOWLEDGE EVALUATION REPORT

        ## MODEL (SCORE)
        instructlab/granite-7b-lab (0.5/1.0)

        ### SCORES (0.0 to 1.0):
        task1 - 0.1
        task2 - 0.9
        """
    )
    assert result.output == expected


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.model.evaluate.launch_server",
    return_value=(mock.MagicMock(), "http://127.0.0.1:8000/v1", 1),
)
@patch(
    "instructlab.eval.mmlu.MMLUBranchEvaluator.run",
    side_effect=[
        (0.5, gen_individual_scores(True)),
        (0.6, gen_individual_scores(False)),
    ],
)
def test_evaluate_mmlu_branch(
    run_mock, launch_server_mock, validate_model_mock, cli_runner: CliRunner
):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mmlu_branch",
            "--model",
            "instructlab/granite-7b-lab",
            "--base-model",
            "instructlab/granite-7b-lab",
            "--tasks-dir",
            "generated",
        ],
    )
    assert validate_model_mock.call_count == 2
    assert launch_server_mock.call_count == 2
    assert run_mock.call_count == 2
    assert result.exit_code == 0
    expected = textwrap.dedent(
        """\
        # KNOWLEDGE EVALUATION REPORT

        ## BASE MODEL (SCORE)
        instructlab/granite-7b-lab (0.6/1.0)

        ## MODEL (SCORE)
        instructlab/granite-7b-lab (0.5/1.0)

        ### IMPROVEMENTS (0.0 to 1.0):
        1. task1: 0.1 -> 0.2 (+0.1)
        2. task3: 0.1 -> 0.2 (+0.1)

        ### REGRESSIONS (0.0 to 1.0):
        1. task2: 0.4 -> 0.3 (-0.1)
        2. task4: 0.4 -> 0.3 (-0.1)

        ### NO CHANGE (0.0 to 1.0):
        1. task5 (0.5)
        """
    )
    assert result.output == expected


@patch("instructlab.model.evaluate.validate_model")
def test_no_model_mt_bench(_, cli_runner: CliRunner):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench",
        ],
    )
    assert (
        result.output
        == "Benchmark mt_bench requires the following args to be set: ['model', 'judge-model']\n"
    )
    assert result.exit_code != 0


def test_missing_benchmark(cli_runner: CliRunner):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
        ],
    )
    assert "Missing option '--benchmark'" in result.output
    assert result.exit_code != 0


def test_invalid_model_mt_bench(cli_runner: CliRunner):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench",
            "--model",
            "invalid",
            "--judge-model",
            "instructlab/merlinite-7b-lab",
        ],
    )
    assert "Model could not be found at 'invalid'" in result.output
    assert result.exit_code != 0


@patch("instructlab.model.evaluate.validate_model")
def test_invalid_max_workers(_, cli_runner: CliRunner):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench",
            "--model",
            "instructlab/granite-7b-lab",
            "--judge-model",
            "instructlab/merlinite-7b-lab",
            "--max-workers",
            "invalid",
        ],
    )
    assert (
        "max-workers must be specified as a positive integer or 'auto'" in result.output
    )
    assert result.exit_code != 0


def test_invalid_model_mmlu(cli_runner: CliRunner):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mmlu",
            "--model",
            "invalid",
        ],
    )
    assert "Model could not be found at 'invalid'" in result.output
    assert result.exit_code != 0


def test_invalid_gguf_model_mmlu(cli_runner: CliRunner):
    with open("model.gguf", "w", encoding="utf-8") as gguf_file:
        gguf_file.write("")
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mmlu",
            "--model",
            "model.gguf",
        ],
    )
    assert (
        "MMLU and MMLUBranch can currently only be used with a safetensors directory"
        in result.output
    )
    assert result.exit_code != 0


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.eval.mmlu.MMLUEvaluator",
    side_effect=Exception("Exiting to check call_args"),
)
def test_int_batchsize_mmlu(mmlu_mock, _, cli_runner: CliRunner):
    cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mmlu",
            "--model",
            "instructlab/granite-7b-lab",
            "--batch-size",
            "1",
        ],
    )
    assert mmlu_mock.call_args_list[0][1]["batch_size"] == 1


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.model.evaluate.launch_server",
    return_value=(mock.MagicMock(), "http://127.0.0.1:8000/v1", 1),
)
def test_invalid_taxonomy_mt_bench_branch(launch_server_mock, _, cli_runner: CliRunner):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench_branch",
            "--model",
            "instructlab/granite-7b-lab",
            "--judge-model",
            "instructlab/merlinite-7b-lab",
            "--base-model",
            "instructlab/granite-7b-lab",
            "--branch",
            "rc",
            "--base-branch",
            "main",
            "--taxonomy-path",
            "invalid_taxonomy",
        ],
    )
    launch_server_mock.assert_called_once()
    assert "Taxonomy git repo not found" in result.output
    assert result.exit_code != 0


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.model.evaluate.launch_server",
    return_value=(mock.MagicMock(), "http://127.0.0.1:8000/v1", 1),
)
def test_invalid_branch_mt_bench_branch(launch_server_mock, _, cli_runner: CliRunner):
    setup_taxonomy("taxonomy")
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench_branch",
            "--model",
            "instructlab/granite-7b-lab",
            "--judge-model",
            "instructlab/merlinite-7b-lab",
            "--base-model",
            "instructlab/granite-7b-lab",
            "--branch",
            "invalid",
            "--base-branch",
            "main",
            "--taxonomy-path",
            "taxonomy",
        ],
    )
    launch_server_mock.assert_called_once()
    assert "Invalid git branch" in result.output
    assert result.exit_code != 0


@patch("instructlab.model.evaluate.validate_model")
@patch(
    "instructlab.model.evaluate.launch_server",
    return_value=(mock.MagicMock(), "http://127.0.0.1:8000/v1", 1),
)
def test_invalid_tasks_dir(_, __, cli_runner: CliRunner):
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mmlu_branch",
            "--model",
            "instructlab/granite-7b-lab",
            "--base-model",
            "instructlab/granite-7b-lab",
            "--tasks-dir",
            "invalid",
        ],
    )
    assert "Tasks dir not found:" in result.output
    assert result.exit_code != 0


def test_invalid_model_path_mmlu(cli_runner: CliRunner, tmp_path):
    test_dir = tmp_path / "test"
    test_dir.mkdir()
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mmlu",
            "--model",
            test_dir,
        ],
    )
    assert (
        "Evaluate '--model' needs to be passed either a safetensors directory or a GGUF file"
        in result.output
    )
    assert result.exit_code != 0


def test_invalid_model_path_mt_bench(cli_runner: CliRunner, tmp_path):
    test_dir = tmp_path / "test"
    test_dir.mkdir()
    result = cli_runner.invoke(
        lab.ilab,
        [
            "--config=DEFAULT",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench",
            "--model",
            test_dir,
        ],
    )
    assert (
        "Evaluate '--model' needs to be passed either a safetensors directory or a GGUF file"
        in result.output
    )
    assert result.exit_code != 0


@patch("instructlab.model.evaluate.validate_model")
def test_vllm_args_null(_, cli_runner: CliRunner):
    fname = common.setup_gpus_config(section_path="serve", vllm_args=lambda: None)
    args = common.vllm_setup_test(
        cli_runner,
        [
            f"--config={fname}",
            "model",
            "evaluate",
            "--benchmark",
            "mt_bench",
            "--model",
            "instructlab/granite-7b-lab",
            "--judge-model",
            "instructlab/merlinite-7b-lab",
            "--gpus",
            "4",
        ],
    )
    common.assert_tps(args, "4")



# Standard
from unittest import mock
import pathlib

# Third Party
import yaml

# First Party
from instructlab import configuration, lab

_CFG_FILE_NAME = "test-serve-config.yaml"


def setup_gpus_config(section_path="serve", gpus=None, tps=None, vllm_args=lambda: []):
    """Returns the name of the config file with the requested vllm config."""
    cfg = configuration.get_default_config()

    section = None
    for subpath in section_path.split("."):
        section = getattr(section or cfg, subpath)

    section.vllm.vllm_args = vllm_args()
    if gpus:
        section.vllm.gpus = gpus
    if tps:
        section.vllm.vllm_args.extend(["--tensor-parallel-size", str(tps)])

    # TODO: generate the name at random?
    with pathlib.Path(_CFG_FILE_NAME).open("w", encoding="utf-8") as f:
        yaml.dump(cfg.model_dump(), f)

    return _CFG_FILE_NAME


@mock.patch("instructlab.model.backends.vllm.check_api_base", return_value=False)
# ^ mimic server *not* running already
@mock.patch(
    "instructlab.model.backends.backends.determine_backend",
    return_value=("vllm", "testing"),
)
@mock.patch("time.sleep", side_effect=Exception("Intended Abort"))
@mock.patch("subprocess.Popen")
def vllm_setup_test(runner, args, mock_popen, *_mock_args):
    mock_process = mock.MagicMock()
    mock_popen.return_value = mock_process

    mock_process.communicate.return_value = ("out", "err")
    mock_process.returncode = 0

    result = runner.invoke(lab.ilab, args)

    if result.exit_code != 0:
        print(result.output)

    assert len(mock_popen.call_args_list) == 1 or (
        "Retrying (1/1)" in result.output and len(mock_popen.call_args_list) == 2
    )

    return mock_popen.call_args_list[0][1]["args"]


def assert_tps(args, tps):
    assert args[-2] == "--tensor-parallel-size"
    assert args[-1] == tps



# SPDX-License-Identifier: Apache-2.0
# pylint: disable=duplicate-code

# Standard

# Third Party
from openai import OpenAI, OpenAIError
import httpx

# Local
from .configuration import DEFAULTS


class ClientException(Exception):
    """An exception raised when invoking client operations."""


def list_models(
    api_base,
    api_key=DEFAULTS.API_KEY,
    http_client=None,
):
    """List models from OpenAI-compatible server"""
    try:
        client = OpenAI(
            base_url=api_base,
            api_key=api_key,
            timeout=DEFAULTS.CONNECTION_TIMEOUT,
            http_client=http_client,
        )
        return client.models.list()
    except OpenAIError as exc:
        raise ClientException(f"Connection Error {exc}") from exc


def check_api_base(api_base: str, http_client: httpx.Client | None = None) -> bool:
    try:
        list_models(api_base=api_base, http_client=http_client)
        return True
    except ClientException:
        return False



# SPDX-License-Identifier: Apache-2.0

# Standard
import glob
import json

# Third Party
import pytest


# This loads each notebook and attempts to parse the JSON that is contained
# within it in order to validate that it is well-formed JSON. This is intended
# to be a smoke test for notebooks until there is a better defined test story
# for them.
@pytest.mark.parametrize("path", glob.glob("**/*.ipynb"))
def test_notebooks(path):
    with open(path, encoding="utf-8") as notebook_file:
        json.load(notebook_file)



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
import abc
import logging
import os
import re
import subprocess

# Third Party
from huggingface_hub import hf_hub_download, list_repo_files
from huggingface_hub import logging as hf_logging
from huggingface_hub import snapshot_download
from packaging import version
import click

# First Party
from instructlab import clickext
from instructlab.configuration import DEFAULTS
from instructlab.utils import is_huggingface_repo, is_oci_repo, load_json

logger = logging.getLogger(__name__)


class Downloader(abc.ABC):
    """Base class for a downloading backend"""

    def __init__(
        self,
        ctx,
        repository: str,
        release: str,
        download_dest: str,
    ) -> None:
        self.ctx = ctx
        self.repository = repository
        self.release = release
        self.download_dest = download_dest

    @abc.abstractmethod
    def download(self) -> None:
        """Downloads model from specified repo/release and stores it into download_dest"""


class HFDownloader(Downloader):
    """Class to handle downloading safetensors and GGUF models from Hugging Face"""

    def __init__(
        self,
        repository: str,
        release: str,
        download_dest: str,
        filename: str,
        hf_token: str,
        ctx,
    ) -> None:
        super().__init__(
            ctx=ctx, repository=repository, release=release, download_dest=download_dest
        )
        self.filename = filename
        self.hf_token = hf_token

    def download(self):
        """
        Download specified model from Hugging Face
        """
        click.echo(
            f"Downloading model from Hugging Face: {self.repository}@{self.release} to {self.download_dest}..."
        )

        if self.hf_token == "" and "instructlab" not in self.repository:
            raise ValueError(
                """HF_TOKEN var needs to be set in your environment to download HF Model.
                Alternatively, the token can be passed with --hf-token flag.
                The HF Token is used to authenticate your identity to the Hugging Face Hub."""
            )

        try:
            if self.ctx.obj is not None:
                hf_logging.set_verbosity(self.ctx.obj.config.general.log_level.upper())
            files = list_repo_files(repo_id=self.repository, token=self.hf_token)
            if any(re.search(r"\.(safetensors|bin)$", fname) for fname in files):
                self.download_entire_hf_repo()
            else:
                self.download_gguf()

        except Exception as exc:
            click.secho(
                f"Downloading model failed with the following Hugging Face Hub error: {exc}",
                fg="red",
            )
            raise click.exceptions.Exit(1)

    def download_gguf(self) -> None:
        try:
            hf_hub_download(
                token=self.hf_token,
                repo_id=self.repository,
                revision=self.release,
                filename=self.filename,
                local_dir=self.download_dest,
            )

        except Exception as exc:
            click.secho(
                f"Downloading GGUF model failed with the following Hugging Face Hub error: {exc}",
                fg="red",
            )
            raise click.exceptions.Exit(1)

    def download_entire_hf_repo(self) -> None:
        try:
            local_dir = os.path.join(self.download_dest, self.repository)
            os.makedirs(name=local_dir, exist_ok=True)

            snapshot_download(
                token=self.hf_token,
                repo_id=self.repository,
                revision=self.release,
                local_dir=local_dir,
            )
        except Exception as exc:
            click.secho(
                f"Downloading safetensors model failed with the following Hugging Face Hub error: {exc}",
                fg="red",
            )
            raise click.exceptions.Exit(1)


class OCIDownloader(Downloader):
    """
    Class to handle downloading safetensors models from OCI Registries
    We are leveraging OCI v1.1 for this functionality
    """

    @staticmethod
    def _extract_sha(sha: str):
        return re.search("sha256:(.*)", sha)

    def _build_oci_model_file_map(self, oci_model_path: str) -> dict:
        """
        Helper function to build a mapping between blob files and what they represent
        Format for the index.json file can be found here: https://github.com/opencontainers/image-spec/blob/main/image-layout.md#indexjson-file
        """
        index_hash = ""
        index_ref_path = f"{oci_model_path}/index.json"
        try:
            index_ref = load_json(Path(index_ref_path))
            match = None
            for manifest in index_ref["manifests"]:
                if (
                    manifest["mediaType"]
                    == "application/vnd.oci.image.manifest.v1+json"
                ):
                    match = self._extract_sha(manifest["digest"])
                    break

            if match:
                index_hash = match.group(1)
            else:
                raise ValueError(
                    f"failed to find hash in the index file: {oci_model_path}"
                )
        except Exception as exc:
            raise ValueError(
                f"failed to extract image hash from index file: {oci_model_path}"
            ) from exc

        blob_dir = f"{oci_model_path}/blobs/sha256"
        index = load_json(Path(f"{blob_dir}/{index_hash}"))

        title_ref = "org.opencontainers.image.title"
        oci_model_file_map = {}
        try:
            for layer in index["layers"]:
                match = self._extract_sha(layer["digest"])

                if match:
                    blob_name = match.group(1)
                    oci_model_file_map[blob_name] = layer["annotations"][title_ref]
        except Exception as exc:
            raise ValueError(
                f"failed to build OCI model file mapping from: {blob_dir}/{index_hash}"
            ) from exc

        return oci_model_file_map

    def download(self):
        click.echo(
            f"Downloading model from OCI registry: {self.repository}@{self.release} to {self.download_dest}..."
        )

        # raise an exception if user specified tag/SHA embedded in repository instead of specifying --release
        match = re.search(r"^(?:[^:]*:){2}(.*)$", self.repository)
        if match:
            click.secho(
                f"Invalid repository supplied: Please specify tag/version '{match.group(1)}' via --release",
                fg="red",
            )
            raise click.exceptions.Exit(1)

        model_name = self.repository.split("/")[-1]
        os.makedirs(os.path.join(self.download_dest, model_name), exist_ok=True)
        oci_dir = f"{DEFAULTS.OCI_DIR}/{model_name}"
        os.makedirs(oci_dir, exist_ok=True)

        # Check if skopeo is installed and the version is at least 1.9
        check_skopeo_version()

        command = [
            "skopeo",
            "copy",
            f"{self.repository}:{self.release}",
            f"oci:{oci_dir}",
            "--remove-signatures",
        ]
        if self.ctx.obj is not None and logger.isEnabledFor(logging.DEBUG):
            command.append("--debug")
        logger.debug(f"Running skopeo command: {command}")

        try:
            subprocess.run(command, check=True, text=True)
        except subprocess.CalledProcessError as e:
            click.secho(
                f"Failed to run skopeo command: {e}. \nstderr: {e.stderr}",
                fg="red",
            )
            raise click.exceptions.Exit(1)

        file_map = self._build_oci_model_file_map(oci_dir)
        if not file_map:
            click.secho(
                "Failed to find OCI image blob hashes",
                fg="red",
            )
            raise click.exceptions.Exit(1)

        blob_dir = f"{oci_dir}/blobs/sha256/"
        for name, dest in file_map.items():
            dest_model_path = Path(self.download_dest) / model_name / str(dest)
            # unlink any existing version of the file
            if dest_model_path.exists():
                dest_model_path.unlink()

            if not dest_model_path.parent.exists():
                dest_model_path.parent.mkdir(parents=True)

            # create symlink to files in cache to avoid redownloading if model has been downloaded before
            os.symlink(
                os.path.join(blob_dir, name),
                dest_model_path,
            )


@click.command()
@click.option(
    "--repository",
    "-rp",
    "repositories",
    multiple=True,
    default=[
        DEFAULTS.MERLINITE_GGUF_REPO,
        DEFAULTS.MISTRAL_GGUF_REPO,
    ],  # TODO: add to config.yaml
    show_default=True,
    help="Hugging Face or OCI repository of the model to download.",
)
@click.option(
    "--release",
    "-rl",
    "releases",
    multiple=True,
    default=[
        "main",
        "main",
    ],  # TODO: add to config.yaml
    show_default=True,
    help="The revision of the model to download - e.g. a branch, tag, or commit hash for Hugging Face repositories and tag or commit hash for OCI repositories.",
)
@click.option(
    "--filename",
    "filenames",
    multiple=True,
    default=[DEFAULTS.GGUF_MODEL_NAME, DEFAULTS.MISTRAL_GGUF_MODEL_NAME],
    show_default="The default model location in the instructlab data directory.",
    help="Name of the model file to download from the Hugging Face repository.",
)
@click.option(
    "--model-dir",
    default=lambda: DEFAULTS.MODELS_DIR,
    show_default="The default system model location store, located in the data directory.",
    help="The local directory to download the model files into.",
)
@click.option(
    "--hf-token",
    default="",
    envvar="HF_TOKEN",
    help="User access token for connecting to the Hugging Face Hub.",
)
@click.pass_context
@clickext.display_params
def download(ctx, repositories, releases, filenames, model_dir, hf_token):
    """Downloads model from a specified repository"""
    downloader = None

    # strict = false ensures that if you just give --repository <some_safetensor> we won't error because len(filenames) is greater due to the defaults
    for repository, filename, release in zip(
        repositories, filenames, releases, strict=False
    ):
        if is_oci_repo(repository):
            downloader = OCIDownloader(
                ctx=ctx, repository=repository, release=release, download_dest=model_dir
            )
        elif is_huggingface_repo(repository):
            downloader = HFDownloader(
                ctx=ctx,
                repository=repository,
                release=release,
                download_dest=model_dir,
                filename=filename,
                hf_token=hf_token,
            )
        else:
            click.secho(
                f"repository {repository} matches neither Hugging Face nor OCI registry format. Please supply a valid repository",
                fg="red",
            )
            raise click.exceptions.Exit(1)

        try:
            downloader.download()
        # pylint: disable=broad-exception-caught
        except (ValueError, Exception) as exc:
            if isinstance(exc, ValueError) and "HF_TOKEN" in str(exc):
                click.secho(
                    f"{downloader.repository} requires a HF Token to be set. Please use '--hf-token' or 'export HF_TOKEN' to download all necessary models",
                    fg="yellow",
                )
            else:
                click.secho(
                    f"Downloading model failed with the following error: {exc}",
                    fg="red",
                )
                raise click.exceptions.Exit(1)


_RECOMMENDED_SCOPEO_VERSION = "1.9.0"


def check_skopeo_version():
    """
    Check if skopeo is installed and the version is at least 1.9.0
    This is required for downloading models from OCI registries.
    """
    # Run the 'skopeo --version' command and capture the output
    try:
        result = subprocess.run(
            ["skopeo", "--version"], capture_output=True, text=True, check=True
        )
    except FileNotFoundError as exc:
        click.secho(
            f"skopeo is not installed, please install recommended version {_RECOMMENDED_SCOPEO_VERSION}",
            fg="red",
        )
        raise click.exceptions.Exit(1) from exc

    logger.debug(f"'skopeo --version' output: {result.stdout}")

    # Extract the version number using a regular expression
    match = re.search(r"skopeo version (\d+\.\d+\.\d+)", result.stdout)
    if match:
        installed_version = match.group(1)
        logger.debug(f"detected skopeo version: {installed_version}")

        # Compare the extracted version with the required version
        if version.parse(installed_version) < version.parse(
            _RECOMMENDED_SCOPEO_VERSION
        ):
            raise ValueError(
                f"skopeo version {installed_version} is lower than {_RECOMMENDED_SCOPEO_VERSION}. Please upgrade it."
            )
    else:
        logger.warning(
            f"Failed to determine skopeo version. Recommended version is {_RECOMMENDED_SCOPEO_VERSION}. Downloading the model might fail."
        )



# SPDX-License-Identifier: MIT

# Standard
from dataclasses import dataclass
import inspect


@dataclass
class BaseModelArgs:
    @classmethod
    def from_dict(cls, params):
        return cls(
            **{
                k: v
                for k, v in params.items()
                if k in inspect.signature(cls).parameters
            }
        )



# SPDX-License-Identifier: Apache-2.0
# pylint: disable=R0801

# Standard
from pathlib import Path
from unittest.mock import patch
import os

# Third Party
from click.testing import CliRunner
import httpx
import pytest

# First Party
from instructlab import lab
from instructlab.model.backends.server import BackendServer, ServerConfig


@pytest.mark.usefixtures("mock_mlx_package")
class TestLabModelTest:
    """Test collection for `ilab model test` command."""

    class ServerMock(BackendServer):
        # py lint: disable=W0613
        def __init__(self):
            sc = ServerConfig("", None)
            super().__init__(
                model_family="",
                model_path="",
                chat_template="",
                host="",
                port=0,
                config=sc,
            )

        def run_detached(
            self,
            http_client: httpx.Client | None = None,
            background: bool = True,
            foreground_allowed: bool = False,
            max_startup_retries: int = 0,
        ) -> str:
            return "api_base_mock"

        def get_backend_type(self):
            return None

        def run(self):
            pass

        def shutdown(self):
            pass

    @patch("instructlab.utils.is_macos_with_m_chip", return_value=False)
    @patch("instructlab.model.linux_test.response", return_value="answer!")
    @patch(
        "instructlab.model.backends.backends.select_backend",
        return_value=ServerMock(),
    )
    def test_model_test_linux(
        self,
        select_backend_mock,
        response_mock,
        is_macos_with_m_chip_mock,
    ):
        runner = CliRunner()
        with runner.isolated_filesystem():
            os.mkdir("generated")
            Path("test_file.jsonl").write_text(
                '{"system": "", "user": "question?", "assistant": ""}', encoding="utf-8"
            )
            result = runner.invoke(
                lab.ilab,
                ["--config=DEFAULT", "model", "test", "--test_file", "test_file.jsonl"],
            )
            assert select_backend_mock.call_count
            assert response_mock.call_count
            assert is_macos_with_m_chip_mock.call_count
            assert "question?" in result.output
            assert "answer!" in result.output
            assert result.exit_code == 0



# SPDX-License-Identifier: Apache-2.0

# Standard

# Third Party
import click

# First Party
from instructlab import clickext
from instructlab.configuration import storage_dirs_exist


@click.group(
    cls=clickext.LazyEntryPointGroup,
    ep_group="instructlab.command.data",
)
@click.pass_context
def data(ctx):
    """Command Group for Interacting with the Data generated by InstructLab.

    If this is your first time running ilab, it's best to start with `ilab config init` to create the environment.
    """
    ctx.obj = ctx.parent.obj
    ctx.obj.ensure_config(ctx)
    if not storage_dirs_exist():
        click.secho(
            "Some ilab storage directories do not exist yet. Please run `ilab config init` before continuing.",
            fg="red",
        )
        raise click.exceptions.Exit(1)
    ctx.default_map = ctx.parent.default_map



# Standard
from unittest.mock import MagicMock
import contextlib
import re

# Third Party
from rich.console import Console
import pytest

# First Party
from instructlab.model.chat import ConsoleChatBot


@pytest.mark.parametrize(
    "model_path,expected_name",
    [
        ("/var/model/file", "file"),
        ("/var/model/directory/", "directory"),
        ("/var/model/directory/////", "directory"),
    ],
)
def test_model_name(model_path, expected_name):
    chatbot = ConsoleChatBot(model=model_path, client=None, loaded={})
    assert chatbot.model_name == expected_name


def handle_output(output):
    return re.sub(r"\s+", " ", output).strip()


def test_list_contexts_output():
    chatbot = ConsoleChatBot(model="/var/model/file", client=None, loaded={})

    def mock_sys_print(output):
        mock_sys_print.output = output

    chatbot._sys_print = mock_sys_print

    mock_prompt_session = MagicMock()
    mock_prompt_session.prompt.return_value = "/lc"
    chatbot.input = mock_prompt_session

    with contextlib.suppress(KeyboardInterrupt):
        chatbot.start_prompt(logger=None)

    console = Console(force_terminal=False)
    with console.capture() as capture:
        console.print(mock_sys_print.output)

    rendered_output = capture.get().strip()

    expected_output = (
        "Available contexts:\n\n"
        "default: I am, Red Hat® Instruct Model based on Granite 7B, an AI language model "
        "developed by Red Hat and IBM Research, based on the Granite-7b-base language "
        "model. My primary function is to be a chat assistant.\n\n"
        "cli_helper: You are an expert for command line interface and know all common "
        "commands. Answer the command to execute as it without any explanation."
    )

    assert handle_output(rendered_output) == handle_output(expected_output)



# SPDX-License-Identifier: Apache-2.0

# Standard
from pathlib import Path
from typing import List
import logging

# Third Party
import click

# First Party
from instructlab import clickext
from instructlab.configuration import DEFAULTS
from instructlab.data.list_data import list_data  # type: ignore
from instructlab.utils import print_table

logger = logging.getLogger(__name__)


@click.command(name="list")
@clickext.display_params
@click.option(
    "--dataset-dirs",
    help="Base directories where datasets are stored.",
    multiple=True,
    default=[DEFAULTS.DATASETS_DIR],
    show_default=True,
)
def list_datasets(dataset_dirs):
    """lists datasets"""

    data: List[List[Path]] = []

    dirs = [Path(dir) for dir in dataset_dirs]

    try:
        data = list_data(dirs)
    except OSError as exc:
        click.secho(f"Failed to list datasets with exception: {exc}")
        raise click.exceptions.Exit(1)

    headers = ["Dataset", "Created At", "File size"]
    print_table(headers, data)



# SPDX-License-Identifier: Apache-2.0

# Standard
from glob import glob
from pathlib import Path
import logging
import os
import shutil

# Third Party
from huggingface_hub import errors as hf_errors
from requests import exceptions as requests_exceptions
import click

# First Party
from instructlab import clickext, utils

logger = logging.getLogger(__name__)


@click.command()
@click.option(
    "--model-dir",
    help="Base directory where model is stored.",
    show_default=True,
    required=True,
)
@click.option("--adapter-file", help="LoRA adapter to fuse.", default=None)
@click.option(
    "-sd",
    "--skip-de-quantize",
    help="Skip de-quantization.",
    is_flag=True,
)
@click.option(
    "-sq",
    "--skip-quantize",
    is_flag=True,
    help="Whether to skip quantization while converting to GGUF.",
)
@click.option(
    "--model-name",
    help="Name of the model being trained/converted. Informs the naming of the final trained model file",
    default=None,
    show_default=True,
)
@click.pass_context
@clickext.display_params
@utils.macos_requirement(echo_func=click.secho, exit_exception=click.exceptions.Exit)
def convert(
    ctx,  # pylint: disable=unused-argument
    model_dir,
    adapter_file,
    skip_de_quantize,
    skip_quantize,
    model_name,
):
    """Converts model to GGUF"""
    # pylint: disable=import-outside-toplevel
    # Third Party
    from instructlab_quantize import run_quantize  # pylint: disable=import-error

    # Local
    from ..llamacpp.llamacpp_convert_to_gguf import convert_llama_to_gguf
    from ..train.lora_mlx.convert import convert_between_mlx_and_pytorch
    from ..train.lora_mlx.fuse import fine_tune

    model_dir = os.path.expandvars(os.path.expanduser(model_dir))

    # compute model name from model-dir if not supplied
    if model_name is None:
        mlx_q_suffix = "-mlx-q"
        model_name = model_dir.split("/")[-1]
        model_name = model_name.replace(mlx_q_suffix, "")

    if adapter_file is None:
        adapter_file = os.path.join(model_dir, "adapters.npz")
    source_model_dir = model_dir
    model_dir_fused = f"{source_model_dir}-fused"

    # this combines adapter with the original model to produce the updated model
    try:
        fine_tune(
            model=source_model_dir,
            save_path=model_dir_fused,
            adapter_file=adapter_file,
            de_quantize=not skip_de_quantize,
        )
    except (requests_exceptions.HTTPError, hf_errors.HFValidationError) as e:
        click.secho(
            f"Failed to fine tune: {e}",
            fg="red",
        )
        raise click.exceptions.Exit(1)

    model_dir_fused_pt = f"{model_name}-trained"
    # this converts MLX to PyTorch
    convert_between_mlx_and_pytorch(
        hf_path=model_dir_fused, mlx_path=model_dir_fused_pt, local=True, to_pt=True
    )

    logger.info(f"deleting {model_dir_fused}...")
    shutil.rmtree(model_dir_fused)

    convert_llama_to_gguf(
        model=Path(model_dir_fused_pt),
        pad_vocab=True,
        skip_unknown=True,
        outfile=f"{model_dir_fused_pt}/{model_name}.gguf",
    )

    logger.info(f"deleting safetensors files from {model_dir_fused_pt}...")
    for file in glob(os.path.join(model_dir_fused_pt, "*.safetensors")):
        os.remove(file)

    # quantize to 4-bit GGUF (optional)
    if not skip_quantize:
        gguf_model_dir = f"{model_dir_fused_pt}/{model_name}.gguf"
        gguf_model_q_dir = f"{model_dir_fused_pt}/{model_name}-Q4_K_M.gguf"
        run_quantize(gguf_model_dir, gguf_model_q_dir, "Q4_K_M")

    logger.info(f"deleting {model_dir_fused_pt}/{model_name}.gguf...")
    os.remove(os.path.join(model_dir_fused_pt, f"{model_name}.gguf"))

    logger.info(f"deleting {source_model_dir}...")
    shutil.rmtree(source_model_dir)



# SPDX-License-Identifier: Apache-2.0

# Standard
from importlib import metadata
import json
import pathlib
import re
import subprocess
import sys
import typing

# Third Party
from click.testing import CliRunner
import click
import pytest

# First Party
from instructlab import configuration as config
from instructlab import lab
from instructlab.clickext import ConfigOption, get_default_and_description
from instructlab.utils import is_macos_with_m_chip


def get_commands():
    commands: typing.Dict[str, typing.Set[str]] = {}
    for primary in metadata.entry_points(group="instructlab.command"):
        sub = commands.setdefault(primary.name, set())
        sub.add(
            ""
        )  # Add empty string to handle primary command without secondary command
        for secondary in metadata.entry_points(
            group=f"instructlab.command.{primary.name}"
        ):
            sub.add(secondary.name)
    return commands


class TestConfig:
    def test_cli_params_hyphenated(self):
        flag_pattern = re.compile("-{1,2}[0-9a-z-]+")
        invalid_flags = []
        commands = get_commands()
        for prim, secondaries in commands.items():
            entry_points = metadata.entry_points(group=f"instructlab.command.{prim}")
            for sec in secondaries:
                for entry_point in entry_points:
                    if entry_point.name != sec:
                        continue
                    command = entry_point.load()
                    if not isinstance(command, click.Command):
                        continue
                    for param in command.params:
                        if not isinstance(param, ConfigOption):
                            continue
                        # Do this long one-liner to avoid linter complaint:
                        # Too many nested blocks (6/5) (too-many-nested-blocks)
                        invalid_flags.extend(
                            f"{prim} {opt}"
                            for opt in param.opts
                            if not flag_pattern.fullmatch(opt)
                        )
        assert not invalid_flags, "<- these commands are using non-hyphenated params"


def test_llamap_cpp_import():
    # Third Party
    import llama_cpp

    llama_cpp.llama_backend_init()


def test_import_mlx():
    # smoke test to verify that mlx is always available on Apple Silicon
    # but never on Linux and Intel macOS.
    if is_macos_with_m_chip():
        assert __import__("mlx")
    else:
        with pytest.raises(ModuleNotFoundError):
            __import__("mlx")


def test_ilab_cli_imports(testdata_path: pathlib.Path):
    script = testdata_path / "leanimports.py"
    subprocess.check_call([sys.executable, str(script)], text=True)


class Command(typing.NamedTuple):
    args: tuple[str, ...]
    extra_args: tuple[str, ...] = ()
    needs_config: bool = True
    should_fail: bool = True

    def get_args(self, *extra, default_config: bool = True) -> list[str]:
        args = []
        if self.needs_config and default_config:
            args.extend(["--config", "DEFAULT"])
        args.extend(self.args)
        args.extend(self.extra_args)
        args.extend(extra)
        return args

    @property
    def has_debug_params(self) -> bool:
        # only second level subcommands have a --debug-params option
        return len(self.args) > 1


subcommands: list[Command] = [
    # split first and second level for nicer pytest output
    # first, second, extra args
    Command((), needs_config=False, should_fail=False),
    Command(("config",), needs_config=False, should_fail=False),
    Command(("config", "edit")),
    Command(("config", "init"), needs_config=False, should_fail=False),
    Command(("config", "show")),
    Command(("model",), needs_config=False, should_fail=False),
    Command(("model", "chat")),
    Command(("model", "convert"), ("--model-dir", "test")),
    Command(("model", "download")),
    Command(("model", "evaluate"), ("--benchmark", "mmlu")),
    Command(("model", "serve")),
    Command(("model", "test")),
    Command(("model", "train")),
    Command(("model", "list")),
    Command(("data",), needs_config=False, should_fail=False),
    Command(("data", "generate")),
    Command(("data", "list")),
    Command(("system",), needs_config=False, should_fail=False),
    Command(("system", "info"), needs_config=False, should_fail=False),
    Command(("taxonomy",), needs_config=False, should_fail=False),
    Command(("taxonomy", "diff")),
]

aliases = [
    "serve",
    "train",
    "chat",
    "generate",
]


@pytest.mark.parametrize("command", subcommands, ids=lambda sc: repr(sc.args))
def test_ilab_cli_help(command: Command, cli_runner: CliRunner):
    cmd = command.get_args("--help")
    assert "--help" in cmd
    result = cli_runner.invoke(lab.ilab, cmd)
    assert result.exit_code == 0, result.stdout


def test_cli_help_matches_field_description(cli_runner: CliRunner):
    commands = get_commands()
    for prim, secondaries in commands.items():
        if prim == "config":
            continue
        for sec in secondaries:
            command_name = f"{prim} {sec}".strip()
            entry_points = metadata.entry_points(group=f"instructlab.command.{prim}")
            for entry_point in entry_points:
                if entry_point.name != sec:
                    continue
                command = entry_point.load()
                if not isinstance(command, click.Command):
                    continue
                command_name_to_list = command_name.split()
                result = cli_runner.invoke(
                    lab.ilab,
                    ["--config", "DEFAULT"] + command_name_to_list + ["--help"],
                )
                assert "Usage:" in result.output
                assert result.exit_code == 0, result.output

                for param in command.params:
                    if not isinstance(param, ConfigOption):
                        continue
                    cfg = config.get_default_config()
                    # Build the config name to retrieve the description
                    # We only want the current command not the parent command
                    # for instance command_name_to_list is "model generate", "model" is
                    # not known in the Pydantic Config so we only want "generate". We
                    # use the last element in case we add more positional arguments in
                    # the future.
                    config_identifier = (
                        [command_name_to_list[-1]] if command_name_to_list[-1] else []
                    )
                    if hasattr(param, "config_sections") and param.config_sections:
                        config_identifier += param.config_sections
                    config_identifier.append(str(param.name))

                    # Fetch the description
                    description, _ = get_default_and_description(cfg, config_identifier)

                    # This is annoying but for additional_args, the description
                    # string is "Additional arguments to pass to the training
                    # script. These arguments are passed as key-value pairs to the
                    # training script." but when the help is generated, the text is
                    # truncated at the key-value word and we end up with
                    # "key-\nvalue", so when we normalize the output and remove the
                    # newlines this gives us "key- value" which is NOT what we want.
                    # So we need to replace "- " with "-" to make sure we match.
                    normalize_output = " ".join(result.stdout.split()).replace(
                        "- ", "-"
                    )
                    assert str(description) in normalize_output, normalize_output


@pytest.mark.parametrize(
    "command",
    [sc for sc in subcommands if sc.has_debug_params],
    ids=lambda sc: repr(sc.args),
)
def test_ilab_cli_debug_params(command: Command, cli_runner: CliRunner):
    result = cli_runner.invoke(lab.ilab, command.get_args("--debug-params"))
    assert result.exit_code == 0, result.stdout

    result = cli_runner.invoke(lab.ilab, command.get_args("--debug-params-json"))
    assert result.exit_code == 0, result.stdout
    j = json.loads(result.stdout)
    assert isinstance(j, dict)


def test_ilab_commands_tested():
    ilab_commands = {None: set([""])}
    for primary in metadata.entry_points(group="instructlab.command"):
        sub = ilab_commands.setdefault(primary.name, set())
        sub.add("")
        for secondary in metadata.entry_points(
            group=f"instructlab.command.{primary.name}"
        ):
            sub.add(secondary.name)

    tested = {None: set([""])}
    for command in subcommands:
        if not command.args:
            continue
        sub = tested.setdefault(command.args[0], set())
        if len(command.args) == 2:
            sub.add(command.args[1])
        else:
            sub.add("")

    assert ilab_commands == tested

    ep_aliases = metadata.entry_points(group="instructlab.command.alias")
    assert set(aliases) == ep_aliases.names


@pytest.mark.parametrize(
    "command",
    subcommands,
    ids=lambda sc: repr(sc.args),
)
def test_ilab_missing_config(command: Command, cli_runner: CliRunner) -> None:
    cmd = command.get_args(default_config=False)
    assert "--config" not in cmd
    result = cli_runner.invoke(lab.ilab, cmd)

    if command.needs_config:
        assert result.exit_code == 2, result
        assert "does not exist or is not a readable file" in result.stdout
    else:
        # should fail due to missing dirs
        if command.should_fail:
            assert result.exit_code == 1, result
            assert (
                "Some ilab storage directories do not exist yet. Please run `ilab config init` before continuing."
                in result.stdout
            )
        else:
            assert result.exit_code == 0, result



# Standard
from unittest import mock
from unittest.mock import patch
import json
import os
import pathlib
import socket
import sys

# Third Party
from click.testing import CliRunner
from safetensors.torch import save_file
import pytest
import torch

# First Party
from instructlab import lab
from instructlab.model.backends import backends, common
from instructlab.model.backends.vllm import build_vllm_cmd
from instructlab.utils import is_model_safetensors


# helper function to create dummy valid and invalid safetensor or bin model directories
def create_safetensors_or_bin_model_files(
    model_path: pathlib.Path, model_file_type: str, valid: bool
):
    test_json_dict = {
        "a": 1,
        "b": 2,
        "quantization_config": {"quant_method": "bitsandbytes"},
    }
    json_object = json.dumps(test_json_dict, indent=4)

    for file in ["tokenizer.json", "tokenizer_config.json"]:
        os.makedirs(os.path.dirname(model_path / file), exist_ok=True)
        with open(model_path / file, "a+", encoding="UTF-8") as f:
            f.write(json_object)

    if model_file_type == "safetensors":
        tensors = {
            "tensor1": torch.randn(3, 3),
            "tensor2": torch.randn(5, 5),
        }
        save_file(tensors, model_path / f"model.{model_file_type}")
    else:
        with open(model_path / f"model.{model_file_type}", "a+", encoding="UTF-8"):
            pass
    if valid:
        with open(model_path / "config.json", "a+", encoding="UTF-8") as f:
            f.write(json_object)


def test_free_port():
    host = "localhost"
    port = common.free_tcp_ipv4_port(host)
    # check that port is bindable
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((host, port))


# the test fails because the model_path is not a valid GGUF file
def test_get_backend_auto_detection_fail_not_gguf(tmp_path: pathlib.Path):
    tmp_gguf = tmp_path / "test.gguf"
    # Write a known invalid header
    invalid_header = (
        b"\x00\x00\x00\x00"  # Use a header that will definitely not match "GGUF_MAGIC"
    )
    tmp_gguf.write_bytes(
        invalid_header + bytes([0] * 4093)
    )  # Fill the rest of the file with zeros
    with pytest.raises(ValueError) as exc_info:
        backends.get(tmp_gguf, None)
    assert "is not a GGUF format" in str(exc_info.value)


# this test succeeds because the model_path is a valid GGUF file (is_model_gguf mocked to returns True)
@patch("instructlab.model.backends.backends.is_model_gguf", return_value=True)
def test_get_backend_auto_detection_success_gguf(
    m_is_model_gguf, tmp_path: pathlib.Path
):
    tmp_gguf = tmp_path / "test.gguf"
    backend = backends.get(tmp_gguf, None)
    assert backend == "llama_cpp"
    m_is_model_gguf.assert_called_once_with(tmp_gguf)


# this tests both cases where a valid and invalid safetensors model directory is supplied
@pytest.mark.parametrize(
    "model_dir,model_file_type,expected",
    [
        (
            "valid_safetensors_model_dir",
            "safetensors",
            True,
        ),
        (
            "valid_bin_model_dir",
            "bin",
            True,
        ),
        (
            "invalid_bin_model_dir",
            "other",
            False,
        ),
        (
            "invalid_safetensors_model_dir",
            "safetensors",
            False,
        ),
    ],
)
def test_is_model_safetensors_or_bin_valid(
    model_dir: str, model_file_type: str, expected: bool, tmp_path: pathlib.Path
):
    model_path = tmp_path / model_dir
    create_safetensors_or_bin_model_files(model_path, model_file_type, expected)

    val = is_model_safetensors(model_path)
    assert val == expected


# this test succeeds because the model_path is a valid safetensors directory and the platform is linux so vllm is
# picked
@patch("sys.platform", "linux")
@pytest.mark.parametrize(
    "safetensors_dir,expected",
    [
        (
            "valid_safetensors_model_dir",
            "vllm",
        ),
    ],
)
def test_get_backend_auto_detection_success_vllm_dir(
    safetensors_dir: str, expected: str, tmp_path: pathlib.Path
):
    safetensors_model_path = tmp_path / safetensors_dir
    create_safetensors_or_bin_model_files(safetensors_model_path, "safetensors", True)

    backend = backends.get(safetensors_model_path, None)
    assert backend == expected


# this test fails because the OS is darwin and a directory is passed, only works on Linux
@patch("sys.platform", "darwin")
@pytest.mark.parametrize(
    "safetensors_dir",
    [
        ("valid_safetensors_model_dir"),
    ],
)
def test_get_backend_auto_detection_failed_vllm_dir_darwin(
    safetensors_dir: str, tmp_path: pathlib.Path
):
    with pytest.raises(ValueError) as exc_info:
        safetensors_model_path = tmp_path / safetensors_dir
        create_safetensors_or_bin_model_files(
            safetensors_model_path, "safetensors", True
        )
        backends.get(safetensors_model_path, None)
    assert "Cannot determine which backend to use" in str(exc_info.value)


# this test succeeds even if the auto-detection picked a different backend, we continue with what
# the user requested
@patch(
    "instructlab.model.backends.backends.determine_backend",
    return_value=("vllm", "reason for selection"),
)
def test_get_forced_backend_fails_autodetection(m_determine_backend):
    backend = backends.get("", "llama-cpp")
    assert backend == "llama-cpp"
    m_determine_backend.assert_called_once()


@mock.patch("instructlab.model.backends.vllm.Server")
@mock.patch("instructlab.model.backends.backends.get", return_value=backends.VLLM)
def test_ilab_vllm_args(
    m_backends_get: mock.Mock,
    m_server: mock.Mock,
    tmp_path: pathlib.Path,
    cli_runner: CliRunner,
):
    cmd = [
        "--config",
        "DEFAULT",
        "model",
        "serve",
        "--model-path",
        str(tmp_path),
        "--backend",
        backends.VLLM,
        "--",
        "--enable_lora",
    ]
    result = cli_runner.invoke(lab.ilab, cmd)
    assert result.exit_code == 0, result.stdout
    m_backends_get.assert_called_once_with(tmp_path, backends.VLLM)
    m_server.assert_called_once_with(
        api_base="http://127.0.0.1:8000/v1",
        chat_template=None,
        model_family=None,
        model_path=tmp_path,
        vllm_args=["--enable_lora"],
        host="127.0.0.1",
        port=8000,
        log_file=None,
    )


@pytest.mark.parametrize(
    "log_file_flag, expected_log_file",
    [
        (None, None),
        ("--log-file", pathlib.Path("test.log")),
    ],
)
@mock.patch("instructlab.model.backends.llama_cpp.Server")
@mock.patch("instructlab.model.backends.backends.get", return_value=backends.LLAMA_CPP)
def test_ilab_llama_cpp_args(
    m_backends_get: mock.Mock,
    m_server: mock.Mock,
    cli_runner: CliRunner,
    log_file_flag,
    expected_log_file,
):
    gguf = pathlib.Path("test.gguf")
    cmd = [
        "--config",
        "DEFAULT",
        "model",
        "serve",
        "--model-path",
        str(gguf),
        "--backend",
        backends.LLAMA_CPP,
    ]
    if log_file_flag:
        cmd.extend([log_file_flag, str(expected_log_file)])

    result = cli_runner.invoke(lab.ilab, cmd)
    assert result.exit_code == 0, result.stdout
    m_backends_get.assert_called_once_with(gguf, backends.LLAMA_CPP)
    m_server.assert_called_once_with(
        api_base="http://127.0.0.1:8000/v1",
        model_path=gguf,
        model_family=None,
        host="127.0.0.1",
        port=8000,
        gpu_layers=-1,
        max_ctx_size=4096,
        num_threads=None,
        chat_template=None,
        log_file=expected_log_file,
    )


def test_build_vllm_cmd_with_defaults(tmp_path: pathlib.Path):
    host = "localhost"
    port = 8080
    model_path = pathlib.Path("/path/to/model")
    model_family = ""
    chat_template = tmp_path / "chat_template.jinja2"
    chat_template.touch()
    vllm_args: list[str] = []
    expected_cmd = [
        sys.executable,
        "-m",
        "vllm.entrypoints.openai.api_server",
        "--host",
        host,
        "--port",
        str(port),
        "--model",
        str(model_path),
        "--chat-template",
        str(chat_template),
        "--distributed-executor-backend",
        "mp",
    ]
    cmd, _ = build_vllm_cmd(
        host, port, model_family, model_path, str(chat_template), vllm_args
    )
    assert cmd == expected_cmd


def test_build_vllm_cmd_with_args_provided(tmp_path: pathlib.Path):
    host = "localhost"
    port = 8080
    model_path = pathlib.Path("/path/to/model")
    model_family = ""
    chat_template = tmp_path / "chat_template.jinja2"
    chat_template.touch()
    vllm_args = [
        "--port",
        str(8001),
        "--model=/path/to/other/model",
        "--distributed-executor-backend",
        "ray",
    ]
    expected_cmd = [
        sys.executable,
        "-m",
        "vllm.entrypoints.openai.api_server",
        "--host",
        host,
        "--chat-template",
        str(chat_template),
    ] + vllm_args

    cmd, _ = build_vllm_cmd(
        host, port, model_family, model_path, str(chat_template), vllm_args
    )
    assert cmd == expected_cmd


def test_build_vllm_cmd_with_bnb_quant(tmp_path: pathlib.Path):
    host = "localhost"
    port = 8080
    model_path = tmp_path / "model"
    model_family = ""
    chat_template = tmp_path / "chat_template.jinja2"
    chat_template.touch()
    vllm_args: list[str] = []
    expected_cmd = [
        sys.executable,
        "-m",
        "vllm.entrypoints.openai.api_server",
        "--host",
        host,
        "--port",
        str(port),
        "--model",
        str(model_path),
        "--chat-template",
        str(chat_template),
        "--quantization",
        "bitsandbytes",
        "--load-format",
        "bitsandbytes",
        "--enforce-eager",
        "--distributed-executor-backend",
        "mp",
    ]
    create_safetensors_or_bin_model_files(model_path, "safetensors", True)
    cmd, _ = build_vllm_cmd(
        host, port, model_family, model_path, str(chat_template), vllm_args
    )
    assert cmd == expected_cmd



# SPDX-License-Identifier: Apache-2.0
# Standard
from os import path

# Third Party
from xdg_base_dirs import xdg_cache_home, xdg_config_home, xdg_data_home
import httpx

ILAB_PACKAGE_NAME = "instructlab"
CONFIG_FILENAME = "config.yaml"
CONFIG_VERSION = "1.0.0"

# Model families understood by ilab
MODEL_FAMILIES = {"merlinite", "mixtral"}

# Map model names to their family
MODEL_FAMILY_MAPPINGS = {
    "granite": "merlinite",
}

# Default log format
LOG_FORMAT = "%(levelname)s %(asctime)s %(name)s:%(lineno)d: %(message)s"


class STORAGE_DIR_NAMES:
    ILAB = "instructlab"
    DATASETS = "datasets"
    CHECKPOINTS = "checkpoints"
    OCI = "oci"
    MODELS = "models"
    TAXONOMY = "taxonomy"
    INTERNAL = (
        "internal"  # for storing all ilab-internal files the user doesn't need to see
    )
    CHATLOGS = "chatlogs"
    PHASED = "phased"


class _InstructlabDefaults:
    """
    Class that defines the default paths used by ilab.
    We define them this way so that they can be lazy-loaded and overridden by tests.
    This way, they are defined when they are read instead of when the module is imported.
    """

    # define static defaults up here
    API_KEY = "no_api_key"

    # ILAB_GLOBAL_CONFIG is the environment variable that can be used to override the default config
    # file. When set, the CLI will use the file specified in the environment variable as a sample to
    # generate the default config file.
    ILAB_GLOBAL_CONFIG = "ILAB_GLOBAL_CONFIG"
    ILAB_TRAIN_PROFILE_DIR = "ILAB_TRAIN_PROFILE_DIR"

    # TODO: Consolidate --model and --model-path into one --model-path flag since we always need a path now
    MODEL_NAME_OLD = "merlinite-7b-lab-Q4_K_M"
    MERLINITE_GGUF_REPO = "instructlab/merlinite-7b-lab-GGUF"
    MISTRAL_GGUF_REPO = "TheBloke/Mistral-7B-Instruct-v0.2-GGUF"
    GGUF_MODEL_NAME = "merlinite-7b-lab-Q4_K_M.gguf"
    MISTRAL_GGUF_MODEL_NAME = "mistral-7b-instruct-v0.2.Q4_K_M.gguf"
    MODEL_REPO = "instructlab/granite-7b-lab"
    JUDGE_MODEL_MT = "prometheus-eval/prometheus-8x7b-v2.0"
    TAXONOMY_REPO = "https://github.com/instructlab/taxonomy.git"
    TAXONOMY_BASE = "origin/main"
    MAX_CONTEXT_SIZE = 4096
    # TODO: these constants should be removed, they should not leak out
    NUM_CPUS = 10
    CHUNK_WORD_COUNT = 1000
    CONNECTION_TIMEOUT = httpx.Timeout(timeout=30.0)
    # use spawn start method, fork is not thread-safe
    MULTIPROCESSING_START_METHOD = "spawn"
    SDG_PIPELINE = "full"
    SDG_SCALE_FACTOR = 30

    # When otherwise unknown, ilab uses this as the default family
    MODEL_FAMILY = "merlinite"
    ADDITIONAL_ARGS_DEFAULTS = {
        "learning_rate": 2e-5,
        "warmup_steps": 25,
        "random_seed": 42,
        "node_rank": 0,
        "nnodes": 1,
        "rdzv_id": 123,
        "rdzv_endpoint": "127.0.0.1:12222",
        "deepspeed_cpu_offload_optimizer_ratio": 1,
        "deepspeed_cpu_offload_optimizer_pin_memory": False,
        "lora_alpha": 32,
        "lora_dropout": 0.1,
        "lora_target_modules": ["q_proj", "k_proj", "v_proj", "o_proj"],
    }

    def __init__(self):
        self._reset()

    def _reset(self):
        """
        Utility function which is mostly used for testing purposes to clear the cache.
        Otherwise, all tests will used cached temporary directories and cause errors.
        """
        self._cache_home = path.join(xdg_cache_home(), ILAB_PACKAGE_NAME)
        self._config_dir = path.join(xdg_config_home(), ILAB_PACKAGE_NAME)
        self._data_dir = path.join(xdg_data_home(), ILAB_PACKAGE_NAME)

    @property
    def CHECKPOINTS_DIR(self) -> str:
        return path.join(self._data_dir, STORAGE_DIR_NAMES.CHECKPOINTS)

    @property
    def OCI_DIR(self) -> str:
        return path.join(self._cache_home, STORAGE_DIR_NAMES.OCI)

    @property
    def DATASETS_DIR(self) -> str:
        return path.join(self._data_dir, STORAGE_DIR_NAMES.DATASETS)

    @property
    def CONFIG_FILE(self) -> str:
        return path.join(self._config_dir, CONFIG_FILENAME)

    @property
    def MODELS_DIR(self) -> str:
        return path.join(self._cache_home, STORAGE_DIR_NAMES.MODELS)

    @property
    def DEFAULT_CHAT_MODEL(self) -> str:
        return path.join(self.MODELS_DIR, self.GGUF_MODEL_NAME)

    @property
    def DEFAULT_TEACHER_MODEL(self) -> str:
        return path.join(self.MODELS_DIR, self.MISTRAL_GGUF_MODEL_NAME)

    @property
    def DEFAULT_JUDGE_MODEL(self) -> str:
        return path.join(self.MODELS_DIR, self.JUDGE_MODEL_MT)

    @property
    def TAXONOMY_DIR(self) -> str:
        return path.join(self._data_dir, STORAGE_DIR_NAMES.TAXONOMY)

    @property
    def CHATLOGS_DIR(self) -> str:
        return path.join(self._data_dir, STORAGE_DIR_NAMES.CHATLOGS)

    @property
    def PHASED_DIR(self) -> str:
        return path.join(self._data_dir, STORAGE_DIR_NAMES.PHASED)

    @property
    def INTERNAL_DIR(self) -> str:
        """
        This directory is used for storing all misc. files that the user doesn't need to see.

        For example, during training we output an intermediate dataset with the tokenized
        instructions and the generated responses.
        Usually this gets outputted into /dev/shm, however this may not be an option on every system,
        so we would store it in here as a fall-back.
        """
        return path.join(self._data_dir, STORAGE_DIR_NAMES.INTERNAL)

    @property
    def PROMPT_FILE(self) -> str:
        return path.join(self.INTERNAL_DIR, "prompt.txt")

    @property
    def SEED_FILE(self) -> str:
        return path.join(self.INTERNAL_DIR, "seed_tasks.json")

    @property
    def EVAL_DATA_DIR(self) -> str:
        return path.join(self.INTERNAL_DIR, "eval_data")

    @property
    def TRAIN_CONFIG_DIR(self) -> str:
        return path.join(self.INTERNAL_DIR, "train_configuration")

    @property
    def TRAIN_PROFILE_DIR(self) -> str:
        return path.join(self.TRAIN_CONFIG_DIR, "profiles")

    @property
    def TRAIN_ADDITIONAL_OPTIONS_DIR(self) -> str:
        return path.join(self.TRAIN_CONFIG_DIR, "additional")

    @property
    def TRAIN_ADDITIONAL_OPTIONS_FILE(self) -> str:
        return path.join(self.TRAIN_ADDITIONAL_OPTIONS_DIR, "additional_args.yaml")

    @property
    def TRAIN_DEFAULT_PROFILE(self) -> str:
        return path.join(self.TRAIN_PROFILE_DIR, "default.yaml")

    @property
    def TRAIN_A100_H100_X4_PROFILE(self) -> str:
        return path.join(self.TRAIN_PROFILE_DIR, "A100_H100_x4.yaml")

    @property
    def TRAIN_A100_H100_X8_PROFILE(self) -> str:
        return path.join(self.TRAIN_PROFILE_DIR, "A100_H100_x8.yaml")

    @property
    def TRAIN_A100_H100_X2_PROFILE(self) -> str:
        return path.join(self.TRAIN_PROFILE_DIR, "A100_H100_x2.yaml")

    @property
    def TRAIN_L40_X8_PROFILE(self) -> str:
        return path.join(self.TRAIN_PROFILE_DIR, "L40_x8.yaml")

    @property
    def TRAIN_L40_X4_PROFILE(self) -> str:
        return path.join(self.TRAIN_PROFILE_DIR, "L40_x4.yaml")

    @property
    def TRAIN_L4_X8_PROFILE(self) -> str:
        return path.join(self.TRAIN_PROFILE_DIR, "L4_x8.yaml")


DEFAULTS = _InstructlabDefaults()



# SPDX-License-Identifier: Apache-2.0

# Standard
import logging
import os
import pathlib

# Local
from .defaults import LOG_FORMAT


class CustomFormatter(logging.Formatter):
    def format(self, record):
        if record.filename == os.path.basename(__file__):
            # Replace our current file name with 'serving_backend' since the message is coming from stdout or
            # stderr. Printing this file name is not useful in this case, we need to print where the line is coming from.
            record.filename = "serving_backend"
            record.lineno = 0  # Remove line number, no point of showing it
        return super().format(record)


class LoggerWriter:
    def __init__(self, logger, level):
        self.logger = logger
        self.level = level

    def write(self, message):
        if message.strip():
            self.logger.log(self.level, message)

    def flush(self):
        pass


def add_file_handler_to_logger(
    logger: logging.Logger, log_file: pathlib.Path | None
) -> None:
    """
    Adds a file handler to the specified logger to write log messages to a file.

    This function checks if a file handler for the specified log file already exists in the logger.
    If it does, the function does nothing. Otherwise, it creates a new file handler that appends
    log messages to the specified file and adds it to the logger.

    Args:
        logger (logging.Logger): The logger to which the file handler should be added.
        log_file (pathlib.Path): The path to the log file where log messages should be written.
    """

    if log_file is None:
        return

    log_file_path = os.path.abspath(log_file)
    logger.debug(f"Checking for existing file handler for log file: {log_file}")

    # Check if a file handler for the specified log file already exists
    for handler in logger.handlers:
        if isinstance(handler, logging.FileHandler):
            if handler.baseFilename == log_file_path:
                logger.debug(
                    f"Log file already exists in handler: {log_file}, doing nothing!"
                )
                return
    logger.debug(f"No file handler found for log file: {log_file}, adding one!")

    # Use the "append" mode
    mode = "a"

    # Create file handler and add it to the logger
    file_handler = logging.FileHandler(log_file, mode, encoding="utf-8")
    logger.debug(f"Adding file handler for log file: {log_file}")
    logger.addHandler(file_handler)

    # Create formatter and set it
    formatter = CustomFormatter(LOG_FORMAT)
    file_handler.setFormatter(formatter)
    logger.debug(f"Added file handler for log file: {log_file}")


def configure_logging(
    *,
    log_level: str,
    debug_level: int = 0,
    fmt: str,
) -> None:
    """Configure logging framework"""
    root = logging.getLogger()

    # reset handlers, removes existing stream logger
    for handler in root.handlers[:]:
        root.removeHandler(handler)
        handler.close()

    stream = logging.StreamHandler()
    stream.setFormatter(CustomFormatter(fmt=fmt))
    root.addHandler(stream)

    instructlab_lgr = logging.getLogger("instructlab")
    openai_lgr = logging.getLogger("openai")
    httpx_lgr = logging.getLogger("httpx")

    # reset loggers to inherit from root logger
    for lgr in [instructlab_lgr, openai_lgr, httpx_lgr]:
        lgr.setLevel(logging.NOTSET)

    if log_level == "DEBUG":
        if debug_level >= 2:
            # debug level 2 enables debug logging for everything
            root.setLevel(logging.DEBUG)
        else:
            # debug level 1 enables debugging for instructlab
            # other loggers are set to INFO (except for openai and httpx)
            root.setLevel(logging.INFO)
            instructlab_lgr.setLevel(logging.DEBUG)
    else:
        # INFO, WARNING, ERROR, FATAL and higher
        root.setLevel(log_level)

    if log_level != "DEBUG" or debug_level < 2:
        # Set logging level of OpenAI client and httpx library to ERROR to
        # suppress INFO messages unless debug_level is >= 2.
        openai_lgr.setLevel(logging.ERROR)
        httpx_lgr.setLevel(logging.ERROR)



# SPDX-License-Identifier: Apache-2.0

documents = [
    """Knowledge is an awareness of facts, a familiarity with individuals and situations,
      or a practical skill. Knowledge of facts, also called propositional knowledge, is often characterized
      as true belief that is distinct from opinion or guesswork by virtue of justification.
      While there is wide agreement among philosophers that propositional knowledge is a form of true belief,
      many controversies focus on justification. This includes questions like how to understand justification,
      whether it is needed at all, and whether something else besides it is needed.
      These controversies intensified in the latter half of the 20th century due to a series of thought experiments
      called Gettier cases that provoked alternative definitions."""
]

knowledge_seed_instruction = [
    {
        "instruction": "What is this knowledge about?",
        "input": "",
        "output": "It's a knowledge that makes the tests more knowledgeable",
        "taxonomy_path": "knowledge->textbook->puns-copy->general",
        "task_description": "For knowledge tests",
        "document": documents,
    },
    {
        "instruction": "question2",
        "input": "",
        "output": "answer2",
        "taxonomy_path": "knowledge->textbook->puns-copy->general",
        "task_description": "For knowledge tests",
        "document": documents,
    },
    {
        "instruction": "question3",
        "input": "",
        "output": "answer3",
        "taxonomy_path": "knowledge->textbook->puns-copy->general",
        "task_description": "For knowledge tests",
        "document": documents,
    },
    {
        "instruction": "question4",
        "input": "",
        "output": "answer4",
        "taxonomy_path": "knowledge->textbook->puns-copy->general",
        "task_description": "For knowledge tests",
        "document": documents,
    },
    {
        "instruction": "Question5",
        "input": "",
        "output": "Answer5",
        "taxonomy_path": "knowledge->textbook->puns-copy->general",
        "task_description": "For knowledge tests",
        "document": documents,
    },
]

generate_data_return_value = [
    {
        "instruction": "3. Tell me a pun about water.",
        "input": "",
        "output": "Why did the scarecrow win an award?\nBecause he was outstanding in his field!",
        "taxonomy_path": "compositional_skills->writing->freeform->jokes->puns-copy->general",
        "task_description": "to teach a large language model to come up with puns",
        "document": None,
    },
    {
        "instruction": "4. Give me a pun about books.",
        "input": "",
        "output": "Why don't books ever get lost on the shelf?\nBecause they are always on the cover!",
        "taxonomy_path": "compositional_skills->writing->freeform->jokes->puns-copy->general",
        "task_description": "to teach a large language model to come up with puns",
        "document": None,
    },
]



# Standard
import enum
import functools
import logging
import os
import pathlib
import pprint
import typing

# Third Party
# pylint: disable=ungrouped-imports
from instructlab.training import DistributedBackend, TorchrunArgs, TrainingArgs
import click

# First Party
from instructlab import utils
from instructlab.configuration import _serve
from instructlab.model.backends import backends

# Local
from .phased_training import (
    EvalPhaseModel,
    EvalResult,
    TrainingJournal,
    TrainingPhases,
    TrainPhaseModel,
)

logger = logging.getLogger(__name__)


class SupportedTrainingStrategies(enum.Enum):
    """Available advanced training stratefies"""

    LAB_MULTIPHASE: str = "lab-multiphase"


def accelerated_train(
    train_args: TrainingArgs,
    torch_args: TorchrunArgs,
    strategy: str | None,
    distributed_backend: str,
    phased_phase1_data: pathlib.Path | None,
    phased_phase2_data: pathlib.Path | None,
    phased_base_dir: pathlib.Path,
    phased_phase1_num_epochs: int | None,
    phased_phase1_samples_per_save: int | None,
    phased_phase1_effective_batch_size: int | None,
    phased_phase2_num_epochs: int | None,
    phased_phase2_samples_per_save: int | None,
    phased_phase2_effective_batch_size: int | None,
    enable_serving_output: bool,
    phased_mt_bench_judge: pathlib.Path | None,
    skip_user_confirm: bool,
    force_clear_phased_cache: bool,
    eval_serve: _serve,
    eval_gpus: int,
    training_journal: pathlib.Path | None,
):
    # run_training is a dynamic attribute, pylint is not clever enough
    # to detect it.
    # Third Party
    if strategy == SupportedTrainingStrategies.LAB_MULTIPHASE.value:
        if (
            distributed_backend
            and distributed_backend not in DistributedBackend._value2member_map_
        ):
            # pylint: disable=broad-exception-raised
            raise Exception(
                f"Invalid training backend option '{distributed_backend}' specified. Please specify either `fsdp` or `deepspeed`"
            )

        # pull the trainrandom.randinting and torch args from the flags
        # the flags are populated from the config as a base.
        logger.debug(
            "Rendered training arguments:\n%s", pprint.pformat(train_args.model_dump())
        )

        if not (phased_phase1_data and phased_phase2_data):
            # pylint: disable=broad-exception-raised
            raise Exception(
                "End-to-end training minimally requires: `--phased-phase1-data`, and `--phased-phase2-data`. One or more wasn't correctly specified."
            )

        # if they both exist, must be Path objects
        if not (phased_phase1_data.exists() and phased_phase2_data.exists()):
            raise FileNotFoundError(
                "Data for both phase1 and phase2 must be specified for phased training."
            )

        mt_bench_judge: pathlib.Path
        if phased_mt_bench_judge is None:
            raise FileNotFoundError(
                "No MT-Bench model was provided with '--phased-mt-bench-judge'"
            )
        if not phased_mt_bench_judge.resolve().exists():
            raise FileNotFoundError(
                f"MT-Bench model directory could not be found: {phased_mt_bench_judge}\nMust be an absolute path to a model directory."
            )

        # makes MyPy happy because 'mt_bench_judge' isn't Path | None
        mt_bench_judge = phased_mt_bench_judge

        if training_journal is None:
            training_journal = phased_base_dir / "journalfile.yaml"
        else:
            # might come in as a str so needs to become a path.
            training_journal = pathlib.Path(training_journal)

        # try to load journal. May be empty.
        journal = TrainingJournal(journalfile=training_journal)
        click.secho("\n\n~~~~~~~~~~~~~STARTING MULTI-PHASE TRAINING~~~~~~~~~~~~~")

        # experimental preference.
        if phased_phase1_num_epochs != 7:
            click.secho(
                f"Running phased training with '{phased_phase1_num_epochs}' epochs.\nNote: 7 epochs is the recommended amount for optimal performance.",
                fg="yellow",
            )

        if journal.was_loaded:
            click.secho(
                f"There was an existing training journal at: '{str(training_journal)}'"
            )
            journal.print_model_rich()
            click.secho(
                f"WARNING: Existing training journal state must correspond to state in '{str(phased_base_dir)}'",
                bg="yellow",
                fg="black",
            )
            click.secho("Alternative behavior is undefined.", bg="yellow", fg="black")
        else:
            click.secho(
                f"No training journal found. Will initialize at: '{str(journal.journalfile)}'",
                bg="yellow",
                fg="black",
            )
            journal.commit(create_new=True)

        user_clear_cache: bool = False
        if not skip_user_confirm:
            click.secho(
                "Metadata (checkpoints, the training journal) may have been saved from a previous training run."
            )
            click.secho(
                "By default, training will resume from this metadata if it exists."
            )
            click.secho(
                "Alternatively, the metadata can be cleared, and training can start from scratch."
            )
            click.secho("\nWould you like to START TRAINING FROM THE BEGINNING?")

            user_clear_cache = click.confirm(
                "'y' clears metadata to start new training, 'N' tries to resume: "
            )

        user_clear_cache = user_clear_cache or force_clear_phased_cache

        if user_clear_cache:
            training_journal.unlink(
                missing_ok=True
            )  # delete whatever the old journal was
            journal = TrainingJournal(
                journalfile=training_journal
            )  # create an empty journal
            journal.commit(create_new=True)  # save it.

        _prepare_phased_base_dir(phased_base_dir, delete_subdirs=user_clear_cache)
        _run_phased_training(
            train_args=train_args,
            torch_args=torch_args,
            base_dir=phased_base_dir,
            phase1_data=phased_phase1_data,
            phase1_num_epochs=phased_phase1_num_epochs,
            phase1_samples_per_save=phased_phase1_samples_per_save,
            phase1_checkpoints_dir=phased_base_dir / "phase1" / "checkpoints",
            phased_phase1_effective_batch_size=phased_phase1_effective_batch_size,
            phase2_data=phased_phase2_data,
            phase2_num_epochs=phased_phase2_num_epochs,
            phase2_samples_per_save=phased_phase2_samples_per_save,
            phase2_checkpoints_dir=phased_base_dir / "phase2" / "checkpoints",
            phased_phase2_effective_batch_size=phased_phase2_effective_batch_size,
            phase2_eval_cache=phased_base_dir / "phase2" / "eval_cache",
            mtbench_judge=mt_bench_judge,
            enable_serving_output=enable_serving_output,
            journal=journal,
            eval_serve=eval_serve,
            eval_gpus=eval_gpus,
        )
    else:
        # Third Party
        from instructlab.training import (
            run_training,  # pylint: disable=no-name-in-module
        )

        try:
            run_training(train_args=train_args, torch_args=torch_args)
        except (RuntimeError, KeyboardInterrupt, Exception) as e:
            if not isinstance(e, KeyboardInterrupt):
                logger.error("Failed during training loop: ", e)
            raise click.exceptions.Exit(1) from e


def _run_phase(
    train_args: TrainingArgs,
    torch_args: TorchrunArgs,
    data_path: pathlib.Path,
    checkpoint_dir: pathlib.Path,
    num_epochs: int | None,
    samples_per_save: int | None,
    effective_batch_size: int | None,
    journal: TrainingJournal,
    phase_model: TrainPhaseModel | EvalPhaseModel,
    next_phase: TrainingPhases,
    model_override: pathlib.Path | None = None,
):
    """Runs a single phase of the multi-phase training pipeline and capture any errors.

    Raises:
        click.exceptions.Exit: Raises a click exception in lieu of a system exit.
    """
    exception = None
    try:
        _training_phase(
            train_args=train_args.model_copy(deep=True),
            torch_args=torch_args,
            data_path=data_path,
            checkpoint_dir=checkpoint_dir,
            num_epochs=num_epochs,
            samples_per_save=samples_per_save,
            effective_batch_size=effective_batch_size,
            model_override=model_override,
            # model override not necessary because we expect model to come from ctx.params.model_path.
        )
    # pylint: disable=broad-except # training library emits a general Exception in lieu of a more specific one.
    except (RuntimeError, KeyboardInterrupt, Exception) as e:
        if not isinstance(e, KeyboardInterrupt):
            logger.error("Failed during training loop: ", e)
        exception = e
    else:
        # progress onto the next phase on success
        journal.current_phase = next_phase
    finally:
        phase_model.ended_at_utc = TrainingJournal.now_utc()
        journal.commit()

    if exception:
        raise click.exceptions.Exit(1) from exception


def _run_phased_training(
    train_args: TrainingArgs,
    torch_args: TorchrunArgs,
    base_dir: pathlib.Path,
    phase1_data: pathlib.Path,
    phase1_num_epochs: int | None,
    phase1_samples_per_save: int | None,
    phase1_checkpoints_dir: pathlib.Path,
    phased_phase1_effective_batch_size: int | None,
    phase2_data: pathlib.Path,
    phase2_num_epochs: int | None,
    phase2_samples_per_save: int | None,
    phase2_checkpoints_dir: pathlib.Path,
    phased_phase2_effective_batch_size: int | None,
    phase2_eval_cache: pathlib.Path,
    mtbench_judge: pathlib.Path,
    enable_serving_output: bool,
    journal: TrainingJournal,
    eval_serve: _serve,
    eval_gpus: int,
) -> None:
    if journal.current_phase == TrainingPhases.DONE:
        click.secho(
            "The selected Training Journal suggests that training has finished, with 'current_phase=done' in the journalfile.",
            fg="cyan",
        )
        return

    # make mypy happy
    phase_model: TrainPhaseModel | EvalPhaseModel | None = None

    if journal.current_phase == TrainingPhases.TRAIN1:
        click.secho("Training Phase 1/2...", fg="cyan")

        phase_model = journal.journal.train_1
        if phase_model is None:
            phase_model = TrainPhaseModel(checkpoints=phase1_checkpoints_dir)
            journal.journal.train_1 = phase_model
        journal.commit()

        _run_phase(
            train_args=train_args,
            torch_args=torch_args,
            data_path=phase1_data,
            checkpoint_dir=phase1_checkpoints_dir,
            num_epochs=phase1_num_epochs,
            samples_per_save=phase1_samples_per_save,
            effective_batch_size=phased_phase1_effective_batch_size,
            journal=journal,
            phase_model=phase_model,
            next_phase=TrainingPhases.TRAIN2,
        )

        logger.debug("Finished training #1\n%s", journal.print_model_rich())
    else:
        click.secho("SKIPPING: Training Phase 1/2; already in Journal", fg="cyan")

    # if journal.current_phase == TrainingPhases.EVAL1:
    #     click.secho("MMLU evaluation for Phase 1...", fg="cyan")

    # NOTE: requires hf_format sub-directory. Training sets this up.
    # phase1_checkpoints_dir = phase1_checkpoints_dir / "hf_format"

    # phase_model = journal.journal.eval_1
    # if phase_model is None:
    #     # if it's not None, it already exists and may have 'results', so we shouldn't overwrite it.
    #     phase_model = EvalPhaseModel(
    #         checkpoints=list(phase1_checkpoints_dir.iterdir())
    #     )
    #     journal.journal.eval_1 = phase_model
    # journal.commit()

    # best_checkpoint = _evaluate_dir_of_checkpoints(
    #     eval_func=_mmlu, phase_model=phase_model, journal=journal
    # )

    # phase_model.best_checkpoint = best_checkpoint
    # phase_model.ended_at_utc = TrainingJournal.now_utc()

    #     journal.current_phase = TrainingPhases.TRAIN2
    #     journal.commit()
    #     logger.debug("Finished eval #1\n%s", journal.print_model_rich())
    # else:
    #     click.secho(
    #         "SKIPPING: MMLU evaluation for Phase 1; already in Journal", fg="cyan"
    #     )

    if journal.current_phase == TrainingPhases.TRAIN2:
        click.secho("Training Phase 2/2...", fg="cyan")

        phase_model = journal.journal.train_2
        if phase_model is None:
            phase_model = TrainPhaseModel(checkpoints=phase2_checkpoints_dir)
            journal.journal.train_2 = phase_model
        journal.commit()

        # if journal.journal.eval_1 is None:
        #     raise RuntimeError(
        #         "Training journal field 'eval_1' cannot be None for phase 'train_2'"
        #     )

        # NOTE:
        # this is a recent change, implemented to ignore MMLU. We just look at the checkpoints
        # from the phase 1 training and take the most recent one.
        phase1_checkpoints_dir_hf = phase1_checkpoints_dir / "hf_format"
        if not phase1_checkpoints_dir_hf.exists():
            raise FileNotFoundError(
                f"{phase1_checkpoints_dir_hf} doesn't exist. This likely means that no checkpoints were saved from phase 1."
            )

        phase1_checkpoints = sorted(
            list(phase1_checkpoints_dir_hf.iterdir()),
            reverse=True,
            # XXX(osilkin): This line takes the checkpoint name "samples_NNN" and tells sorted
            #               to use the last NNN string as an integer
            key=lambda x: int(str(x).rsplit("_", maxsplit=1)[-1]),
        )

        if len(phase1_checkpoints) == 0:
            raise FileNotFoundError(
                f"{phase1_checkpoints_dir_hf} Has no checkpoints. This likely means that no checkpoints were saved from phase 1."
            )

        next_checkpoint = phase1_checkpoints[0]
        _run_phase(
            train_args=train_args,
            torch_args=torch_args,
            data_path=phase2_data,
            checkpoint_dir=phase2_checkpoints_dir,
            num_epochs=phase2_num_epochs,
            samples_per_save=phase2_samples_per_save,
            effective_batch_size=phased_phase2_effective_batch_size,
            journal=journal,
            phase_model=phase_model,
            next_phase=TrainingPhases.EVAL2,
            model_override=next_checkpoint,
        )
        logger.debug("Finished training #2\n%s", journal.print_model_rich())
    else:
        click.secho("SKIPPING: Training Phase 2/2; already in Journal", fg="cyan")

    if journal.current_phase == TrainingPhases.EVAL2:
        click.secho("MT-Bench evaluation for Phase 2...", fg="cyan")

        phase2_checkpoints_dir = phase2_checkpoints_dir / "hf_format"
        phase2_eval_cache = base_dir / "phase2" / "eval_cache"
        phase_model = journal.journal.eval_2
        if phase_model is None:
            # if it's not None, it already exists and may have 'results', so we shouldn't overwrite it.
            phase_model = EvalPhaseModel(
                checkpoints=list(phase2_checkpoints_dir.iterdir())
            )
            journal.journal.eval_2 = phase_model
        journal.commit()

        best_checkpoint = _evaluate_dir_of_checkpoints(
            phase_model=phase_model,
            journal=journal,
            eval_func=functools.partial(
                _mtbench,
                eval_serve=eval_serve,
                eval_gpus=eval_gpus,
                eval_cache=phase2_eval_cache,
                mtbench_judge=mtbench_judge,
                enable_serving_output=enable_serving_output,
            ),
        )

        phase_model.best_checkpoint = best_checkpoint
        phase_model.ended_at_utc = TrainingJournal.now_utc()

        journal.current_phase = TrainingPhases.DONE
        journal.journal.final_output = best_checkpoint
        journal.journal.ended_at_utc = TrainingJournal.now_utc()
        journal.commit()
        logger.debug("Finished eval #2\n%s", journal.print_model_rich())

    else:
        click.secho(
            "SKIPPING: MT-Bench evaluation for Phase 2; already in Journal", fg="cyan"
        )

    output_checkpoint: EvalResult | None = journal.journal.final_output
    if not output_checkpoint:
        raise RuntimeError(
            "At the end of training, but no 'final_output' checkpoint in TrainingJournal"
        )

    click.secho(
        f"Training finished! Best final checkpoint: {output_checkpoint.checkpoint} with score: {output_checkpoint.score}\nJournal: {str(journal.journalfile)}",
        fg="green",
    )


def _training_phase(
    train_args: TrainingArgs,
    torch_args: TorchrunArgs,
    data_path: pathlib.Path,
    model_override: pathlib.Path | None = None,
    num_epochs: int | None = None,
    samples_per_save: int | None = None,
    checkpoint_dir: pathlib.Path | None = None,
    effective_batch_size: int | None = None,
) -> None:
    """A single step of phased training that supports key param overriding."""

    # Third Party
    from instructlab.training import run_training  # pylint: disable=no-name-in-module

    logger.debug(
        f"Phased Training -- training phase -- Overriding data_path: {train_args.data_path} with {data_path}"
    )

    # NOTE: have to cast pathlib.Path to str because Pydantic models require this. Here and below.
    train_args.data_path = str(data_path)

    if checkpoint_dir:
        train_args.ckpt_output_dir = str(checkpoint_dir)

    if model_override:
        logger.debug(
            f"Phased Training -- training phase -- Overriding model_path: {train_args.model_path} with {model_override}"
        )
        train_args.model_path = str(model_override)

    if num_epochs:
        logger.debug(
            f"Phased Training -- training phase -- Overriding num epochs: {train_args.num_epochs} with {num_epochs}"
        )
        train_args.num_epochs = num_epochs

    if samples_per_save is not None:
        logger.debug(
            f"Phased Training -- training phase -- Overriding samples per save: {train_args.save_samples} with {samples_per_save}"
        )
        train_args.save_samples = samples_per_save

    if effective_batch_size:
        logger.debug(
            f"Phased Training -- training phase -- Overriding effective batch size: {train_args.effective_batch_size} with {effective_batch_size}"
        )
        train_args.effective_batch_size = effective_batch_size

    click.secho(
        f"TrainingArgs for current phase: {pprint.pformat(train_args)}", fg="cyan"
    )

    run_training(train_args=train_args, torch_args=torch_args)


def _prepare_phased_base_dir(
    phased_base_dir: pathlib.Path, delete_subdirs: bool = True
) -> None:
    """Adds phase1 and phase2 directories in phased_base_dir.
    In each, adds `checkpoints` and `eval_cache` subdirectories.

    Also adds training `journalfile.yaml`

    Args:
        phased_base_dir: directory wrapping phase1 and phase2 cached data.
    """

    logger.debug(f"Phased Training -- Preparing phased base dir: {phased_base_dir}")

    phase1_dir_path = phased_base_dir / "phase1"
    phase2_dir_path = phased_base_dir / "phase2"

    for p in [phase1_dir_path, phase2_dir_path]:
        if delete_subdirs:
            utils.clear_directory(p)
        _setup_phase_dirs(p)


def _setup_phase_dirs(path: pathlib.Path) -> None:
    """Creates {path}/checkpoints and {path}/eval_cache directories."""

    # TODO: these sub-directory names are hard-coded here but they
    # could be parameterized in config.
    logger.debug(f"Phased Training -- Created phase directories for {path}")
    ckpt_path = path / "checkpoints"
    eval_cache_path = path / "eval_cache"

    os.makedirs(ckpt_path, exist_ok=True)
    os.makedirs(eval_cache_path, exist_ok=True)


def _mmlu(model: pathlib.Path) -> float:
    # Third Party
    from instructlab.eval.mmlu import MMLU_TASKS, MMLUEvaluator
    import torch

    tasks = MMLU_TASKS
    if os.environ.get("INSTRUCTLAB_EVAL_MMLU_MIN_TASKS") is not None:
        tasks = tasks[:4]
    evaluator = MMLUEvaluator(model, tasks=tasks)

    # type the variable because MyPy doesn't seem to honor the types of the spread tuple
    ckpt_score: float
    ckpt_score, _ = evaluator.run()

    logging.debug("Phased Training -- MMLU eval phase -- Clearing PyTorch cache")
    torch.cuda.empty_cache()

    return ckpt_score


def _mtbench(
    model: pathlib.Path,
    eval_serve: _serve,
    eval_gpus: int,
    eval_cache: pathlib.Path,
    mtbench_judge: pathlib.Path,
    enable_serving_output: bool,
) -> float:
    # TODO: optimization: run all generations in serial and then do all judgments at once to save time loading/unloading prometheus.
    # Third Party
    from instructlab.eval.mt_bench import MTBenchEvaluator
    import torch

    # First Party
    from instructlab.model.evaluate import get_gpus, get_model_name, launch_server

    explicit_gpus = None
    gpus, effective_gpus = get_gpus(eval_serve, eval_gpus)
    if gpus and gpus > 0:
        # gpus are specified in config for evaluate
        logger.debug("Using gpus from config")
        explicit_gpus = gpus
    elif effective_gpus > 0:
        # tensor-parallel size specified in serving config
        logger.debug("Using gpus from --tensor-parallel-size in config")
    else:
        # TODO: Should be parameterized by the user specific for training
        explicit_gpus = torch.cuda.device_count()
        effective_gpus = explicit_gpus

    model_name = get_model_name(str(model))
    judge_model_name = get_model_name(str(mtbench_judge))

    evaluator = MTBenchEvaluator(
        model_name=model_name,
        judge_model_name=judge_model_name,
        output_dir=str(eval_cache),
        merge_system_user_message=True,  # TODO: expose this to the user
    )

    server = None
    model_serve_url = None
    try:
        logger.debug("Starting model server for mt-bench answer generation")
        server, model_serve_url, effective_gpus = launch_server(
            eval_serve=eval_serve,
            tls_client_cert=None,
            tls_client_key=None,
            tls_client_passwd=None,
            tls_insecure=False,
            model=str(model),
            model_name=model_name,
            gpus=explicit_gpus,
            max_workers="auto",
            enable_serving_output=enable_serving_output,
            backend=backends.VLLM,
        )
        logger.debug("Generating mt-bench answers")
        evaluator.gen_answers(
            model_serve_url, max_workers="auto", serving_gpus=effective_gpus
        )
    finally:
        if server is not None:
            server.shutdown()

    try:
        logger.debug("Starting model server for mt-bench answer judgment")
        server, model_serve_url, effective_gpus = launch_server(
            eval_serve=eval_serve,
            tls_client_cert=None,
            tls_client_key=None,
            tls_client_passwd=None,
            tls_insecure=False,
            model=str(mtbench_judge),
            model_name=judge_model_name,
            gpus=explicit_gpus,
            max_workers="auto",
            backend=backends.VLLM,
            enable_serving_output=enable_serving_output,
        )
        logger.debug("Judging mt-bench answers")
        mt_bench_results: tuple = evaluator.judge_answers(
            model_serve_url, max_workers="auto", serving_gpus=effective_gpus
        )
        ckpt_score: float = mt_bench_results[0]
    finally:
        if server is not None:
            server.shutdown()

    return ckpt_score


def _evaluate_dir_of_checkpoints(
    eval_func: typing.Callable[..., float],
    phase_model: EvalPhaseModel,
    journal: TrainingJournal,
) -> EvalResult:
    """Run eval_func on all model checkpoints in a directory."""
    # TODO: parallelize MMLU over available GPUs

    # doing this to avoid removing checkpoints from same list that we're iterating over.
    checkpoints_todo = list(
        set(phase_model.checkpoints) - set(phase_model.finished_checkpoints)
    )

    if len(checkpoints_todo) == 0:
        raise RuntimeError(
            "No checkpoints were evaluated, 'checkpoints_todo' was empty in journal."
        )

    for checkpoint in checkpoints_todo:
        logger.debug(str(checkpoint))
        checkpoint_score = eval_func(model=checkpoint)

        phase_model.results.append(
            EvalResult(
                score=checkpoint_score,
                checkpoint=checkpoint,
                ended_at_utc=TrainingJournal.now_utc(),
            )
        )

        phase_model.finished_checkpoints.append(checkpoint)
        journal.commit()

        click.secho(
            f"CHECKPOINT EVALUATION: {str(checkpoint)} SCORED {checkpoint_score}",
            fg="red",
            bg="cyan",
        )

    return TrainingJournal.best_checkpoint(phase_model=phase_model)



# Standard
import contextlib
import sys

# Third Party
import yaml

# WARNING:
# This script is a utility meant to be used in scripts/e2e.sh to edit the entries
# in the ilab config from the CLI. It should not be used edit the ilab config by users


def update_yaml(file_path, key, new_value):
    try:
        with open(file_path, "r") as file:
            try:
                data = yaml.safe_load(file)
            except yaml.YAMLError as exc:
                print(f"Error parsing YAML: {exc}")
                sys.exit(1)

        keys = key.split(".")
        values = []

        with contextlib.suppress(ValueError):
            new_value = int(new_value)

        for key in keys:
            if key in data:
                values.append(data)
                data = data[key]
            else:
                print(f"Error: Key '{key}' not found in the YAML file.")
                sys.exit(1)

        if len(keys) != len(values):
            print("Error: Keys and Values list sizes are not equal")
            sys.exit(1)

        keys.reverse()
        values.reverse()
        values[0][keys[0]] = new_value
        for i in range(1, len(keys)):
            values[i][keys[i]] = values[i - 1]
        data = values[len(keys) - 1]

        try:
            with open(file_path, "w") as file:
                try:
                    yaml.safe_dump(data, file)
                except yaml.YAMLError as exc:
                    print(f"Error dumping YAML: {exc}")
                    sys.exit(1)
        except OSError:
            print(f"Error: The file '{file_path}' was not found.")
            sys.exit(1)

        print(f"Updated '{key}' to '{new_value}' in {file_path}")
    except OSError:
        print(f"Error: The file '{file_path}' was not found.")
        sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print(
            "Usage: python config_editor.py <path to yaml config file> <key(s)> <new value>"
        )
        print("Ex: ")
        print("   python config_editor.py config.yaml evaluate.gpus 4")
        print(
            "   python config_editor.py config.yaml generate.teacher.vllm.model_family mixtral"
        )
        sys.exit(1)

    print(
        "\033[31mWARNING: This script is a utility only meant to be used in scripts/e2e.sh. It should not be used by users to edit the ilab config.yaml\033[0m"
    )

    yaml_file = sys.argv[1]
    yaml_keys = sys.argv[2]
    new_value = sys.argv[3]

    update_yaml(yaml_file, yaml_keys, new_value)



# SPDX-License-Identifier: Apache-2.0

SYS_PROMPT = "I am, Red Hat® Instruct Model based on Granite 7B, an AI language model developed by Red Hat and IBM Research, based on the Granite-7b-base language model. My primary function is to be a chat assistant."



# Standard
from copy import copy
from functools import partial
from pathlib import Path
import logging
import math
import os

# Third Party
from instructlab_quantize import run_quantize
from tqdm import tqdm
from transformers import Adafactor, AutoModelForCausalLM
import numpy as np
import psutil

# First Party
from instructlab.llamacpp import llamacpp_convert_to_gguf

logger = logging.getLogger(__name__)


def train(train_args, device):
    """
    train runs a CPU and MacOS optimized version of full fine tuning.
    Adafactor is the optimizer of choice and the multiprocessing method is set to spawn.
    Dataloading functions imported from the training library.
    """

    # pylint: disable=no-name-in-module
    # Third Party
    from instructlab.training import config
    from instructlab.training import data_process as dp
    from instructlab.training import (
        multipack_sampler,
        token_dataset,
        tokenizer_utils,
        utils,
    )
    from torch.utils.data import DataLoader
    import torch

    dp.main(
        config.DataProcessArgs(
            data_output_path=train_args.data_output_dir,
            model_path=train_args.model_path,
            data_path=train_args.data_path,
            max_seq_len=train_args.max_seq_len,
            chat_tmpl_path=train_args.chat_tmpl_path,
        )
    )

    # load chat template based on path in the args
    CHAT_TEMPLATE, SPECIAL_TOKENS = utils.retrieve_chat_template(
        train_args.chat_tmpl_path
    )
    # get the tokenizer for the model
    tokenizer = tokenizer_utils.setup_tokenizer(
        train_args.model_path, SPECIAL_TOKENS, CHAT_TEMPLATE
    )

    # setup the dataset and place it in data.jsonl, this needs to be used for training NOT the jsonl produced by sdg
    dataset = token_dataset.setup_dataset(
        os.path.join(train_args.data_output_dir, "data.jsonl"),
    )

    # based on the length of the dataset, figure out the max batch len
    packing_max_batch_len, accum = (
        multipack_sampler.find_packing_max_batch_len_and_grad_accum(
            num_gpus=1,
            avg_sample_len=dataset.get_lengths().mean(),
            effective_batch_size=train_args.effective_batch_size,
            max_batch_len_per_gpu=train_args.max_batch_len,
            is_padding=False,
            dataset=dataset,
            seed=47,
        )
    )

    collate_fn = partial(pad_collate_fn, pad_token_id=tokenizer.pad_token_id)

    # use a multipack sampler since this is a non-dist. scenario, training library defaults to this as well
    sampler = multipack_sampler.MultipackDistributedBatchSampler(
        batch_max_length=packing_max_batch_len,
        lengths=dataset.get_lengths(),
        num_replicas=1,
        rank=0,
        seed=47,
        padding=True,
    )
    sampler = {"batch_sampler": sampler}

    # 4 workers for the dataloader as compared to the original 8 to optimize performance
    # this dataloader needs a spawn multiproc method to optimize memory and to work on Apple Silicon.
    dataloader = DataLoader(
        dataset,
        **sampler,
        num_workers=4,
        collate_fn=collate_fn,
    )
    dataloader.multiprocessing_context = "spawn"

    logger.info(
        f"avg_sample_len: {dataset.get_lengths().mean()}\n effective_batch_size: {train_args.effective_batch_size}\n max_batch_len: {train_args.max_batch_len}\n packing_max_batch_len: {packing_max_batch_len} \n grad_accum: {accum}\n  num_batches: {len(dataloader)}\n avg_samples_per_batch: {len(dataset) / len(dataloader)}"
    )
    # set device based off argument given
    dev = torch.device(device)
    # auto model based on model path
    model = AutoModelForCausalLM.from_pretrained(train_args.model_path)
    if len(tokenizer) > model.config.vocab_size:
        logger.warning(
            f"WARNING: tokenizer has {len(tokenizer)} tokens but model has {model.config.vocab_size} vocab size"
        )
        model.resize_token_embeddings(
            int(8 * math.ceil(len(tokenizer) / 8.0))
        )  # make the vocab size multiple of 8 for sharding the embedding layer.

    # Fix any discrepancy between model and tokenizer
    # bos and eos tokens mark the beginning and end of a sequence/sentence
    if (
        model.config.pad_token_id is not None
        and tokenizer.pad_token_id is not None
        and model.config.pad_token_id != tokenizer.pad_token_id
    ):
        logger.warning(
            f"There is a mismatch between pad token id of model ({model.config.pad_token_id}) and tokenizer ({tokenizer.pad_token_id}). Fixing model pad token id to be same as tokenizer's pad token id."
        )
        model.config.pad_token_id = tokenizer.pad_token_id
    if (
        model.config.bos_token_id is not None
        and tokenizer.bos_token_id is not None
        and model.config.bos_token_id != tokenizer.bos_token_id
    ):
        logger.warning(
            f"There is a mismatch between bos token id of model ({model.config.bos_token_id}) and tokenizer ({tokenizer.bos_token_id}). These tokens denote the start of a sequence of data. Fixing model bos token id to be same as tokenizer's bos token id."
        )
        model.config.bos_token_id = tokenizer.bos_token_id
    if (
        model.config.eos_token_id is not None
        and tokenizer.eos_token_id
        and model.config.eos_token_id != tokenizer.eos_token_id
    ):
        logger.warning(
            f"There is a mismatch between eos token id of model ({model.config.eos_token_id}) and tokenizer ({tokenizer.eos_token_id}). These tokens denote the end of a sequence of data. Fixing model eos token id to be same as tokenizer's eos token id."
        )
        model.config.eos_token_id = tokenizer.eos_token_id

    model = utils.convert_loss_to_reduce_sum(model)
    model = utils.add_noisy_embeddings(model, noise_alpha=None)

    # Get virtual memory statistics
    memory_info = psutil.virtual_memory()

    # Total RAM
    total_ram = memory_info.total / (1024**3)  # Convert to GB
    logger.info(f"Total RAM: {total_ram:.2f} GB")
    # if RAM is <= 16, we need to use fp16 not fp32. This will yield a worse model but will allow the full pipeline to run
    if total_ram <= 16:
        # if <= 16GB ram, use gradinent accum and hald precision
        logger.warning(
            f"Your system has {total_ram:.2f} GB of RAM. This is below our reccomendation of 32GB for this type of training. Using half precision."
        )
        model = model.to(dev).half()  # Convert model to float16
    else:
        model = model.to(dev)

    # adafactor and gradient checkpointing are memory friendly, we opt to use these in the CPU/MPS loop to fit 7b models.
    optimizer = Adafactor(
        model.parameters(), lr=2e-5, scale_parameter=False, relative_step=False
    )
    model.gradient_checkpointing_enable()

    model.train()

    # For each epoch do the following:
    # 1. Incremement Loading Bar
    # 2. put zeros onto the proper device so we can store some values we need
    # 3. put each part of the batch onto the device as well using fp6
    # 4. Forward pass on the model. If using half, we need to do some special handling. If not, then just put the batch onto the device using fp32
    # 5. Calculate the loss
    # 6. backward pass on the loss
    # 7. We are using gradient accumulation to be memory efficient, so if this step is divisible by 4, take a step and zero out the gradients (freeing memory)
    # 8. clear out some caches if on MPS
    # 9. Incrememnt loading bar
    # 10. at the end og the epoch, save the checpoint to a samples folder and convert to GGUF as well
    for epoch in range(train_args.num_epochs):
        dataloader.batch_sampler.set_epoch(epoch)
        inner_pb = tqdm(range(len(dataloader)), desc=f"Epoch {epoch}")
        aggregated_values = torch.zeros(3, dtype=torch.float16).to(dev)

        for step, batch in enumerate(dataloader):
            aggregated_values[0] = batch.pop("num_loss_counted_tokens")
            aggregated_values[1] = len(batch["input_ids"])

            # Move and cast batch data to device
            for k in batch:
                if total_ram < 16:
                    if k in ["input_ids", "attention_mask"]:
                        # these two need to be of type long if using .half()
                        batch[k] = batch[k].to(device=dev, dtype=torch.long)
                    else:
                        batch[k] = batch[k].to(device=dev, dtype=torch.float16)
                else:
                    batch[k] = batch[k].to(device=dev)

            output = model(**batch, use_cache=False)
            loss = output.loss
            aggregated_values[2] = loss.item()

            num_loss_counted_tokens = aggregated_values[0]
            loss = loss / num_loss_counted_tokens

            loss = loss / accum  # Scale the loss for accumulation steps

            logger.info(f"\nEpoch: {epoch}, Step: {step + 1}, Rank: 0, loss = {loss}")

            # Gradient accumulation
            loss.backward()  # Backward pass

            # Clear cache before optimizer step
            # below we clear the MPS cache quite a bit. MPS caching is different from CPU.
            # on MPS systems with 32 or 64 GB of unified memory, it is important to clear the cache beacuse often we are paging a lot of the data that can't fit directly on GPU
            # CPU doesn't have an equivalent cache clearing method, only GPU devices like MPS, CUDA, ROCM, etc.
            if dev.type == "mps":
                torch.mps.empty_cache()

            # if we are on a step which is divisible by 4, step and zero gradients
            if (step + 1) % accum == 0:
                optimizer.step()  # Optimizer step
                optimizer.zero_grad()  # Zero gradients

                # Clear cache after optimizer step
                if dev.type == "mps":
                    torch.mps.empty_cache()

            inner_pb.update(1)

        # Clear cache at the end of the epoch if needed
        if dev.type == "mps":
            torch.mps.empty_cache()

        output_dir = (
            Path(train_args.ckpt_output_dir) / "hf_format" / f"samples_{(epoch*8)}"
        )

        logger.info(f"Saving Model to: {output_dir}")
        model_state = model.state_dict()
        output_dir.mkdir(parents=True, exist_ok=True)
        output_model_file = output_dir / "pytorch_model.bin"
        output_config_file = output_dir / "config.json"

        tmp_conf = copy(model.config)

        torch.save(model_state, str(output_model_file))
        tmp_conf.to_json_file(str(output_config_file))
        tokenizer.save_pretrained(str(output_dir))

        # convert to GGUF at the end so llamacpp can serve the model originally in .bin
        llamacpp_convert_to_gguf.convert_llama_to_gguf(
            model=output_dir,
            pad_vocab=True,
            skip_unknown=True,
            outfile=os.path.join(output_dir, "pytorch_model.gguf"),
        )

        # quantize the model so everyone can run it
        gguf_model_dir = os.path.join(output_dir, "pytorch_model.gguf")
        gguf_model_q_dir = os.path.join(output_dir, "pytorch_model-Q4_K_M.gguf")
        run_quantize(gguf_model_dir, gguf_model_q_dir, "Q4_K_M")


def pad_collate_fn(batch, pad_token_id):
    lens = np.array([len(item["input_ids"]) for item in batch])
    max_len = max(lens)
    # Third Party
    import torch
    import torch.nn.functional as F

    input_ids = torch.stack(
        [
            F.pad(
                item["input_ids"],
                (max_len - len(item["input_ids"]), 0),
                mode="constant",
                value=pad_token_id,
            )
            for item in batch
        ]
    )
    labels = torch.stack(
        [
            F.pad(
                item["labels"],
                (max_len - len(item["labels"]), 0),
                mode="constant",
                value=-100,
            )
            for item in batch
        ]
    )
    num_loss_counted_tokens = (labels != -100).sum()

    attention_mask = torch.stack(
        [
            F.pad(
                item["attention_mask"],
                (max_len - len(item["attention_mask"]), 0),
                mode="constant",
                value=0,
            )
            for item in batch
        ]
    )
    logger.info(
        f"\033[96m total tokens: {max_len * len(batch)} num samples: {len(batch)} num padding tokens: {max_len * len(batch) - lens.sum()} - rank: {0} "
        f"max len: {max_len} min len: {min(lens)} avg len: {lens.mean()} "
        f"num_loss_counted_tokens: {num_loss_counted_tokens}\033[0m"
    )

    return {
        "input_ids": input_ids,
        "labels": labels,
        "num_loss_counted_tokens": num_loss_counted_tokens,
        "attention_mask": attention_mask,
    }



# SPDX-License-Identifier: Apache-2.0

# Standard
import logging
import pathlib
import signal

# Third Party
import click

# First Party
from instructlab import clickext
from instructlab.model.serve_backend import serve_backend, signal_handler

logger = logging.getLogger(__name__)

# Register the signal handler for SIGTERM
signal.signal(signal.SIGTERM, signal_handler)


def warn_for_unsupported_backend_param(ctx):
    for param in ["gpu_layers", "num_threads", "max_ctx_size"]:
        if ctx.get_parameter_source(param) == click.core.ParameterSource.COMMANDLINE:
            logger.warning(
                f"Option '--{param.replace('_','-')}' not supported by the backend."
            )


@click.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": False},
)
@click.option(
    "--model-path",
    type=click.Path(path_type=pathlib.Path),
    cls=clickext.ConfigOption,
    required=True,  # default from config
)
@click.option(
    "--gpu-layers",
    type=click.INT,
    cls=clickext.ConfigOption,
    config_sections="llama_cpp",
    required=True,  # default from config
)
@click.option(
    "--num-threads",
    type=click.INT,
    required=False,
    help="The number of CPU threads to use.",
)
@click.option(
    "--max-ctx-size",
    type=click.INT,
    cls=clickext.ConfigOption,
    config_sections="llama_cpp",
)
@click.option(
    "--model-family",
    type=str,
    help="Model family is used to specify which chat template to serve with",
)
@click.option(
    "--log-file",
    type=click.Path(path_type=pathlib.Path),
    required=False,
    help="Log file path to write server logs to.",
)
@click.option(
    "--chat-template",
    type=str,
    cls=clickext.ConfigOption,
)
@click.option(
    "--backend",
    type=click.Choice(("llama_cpp", "vllm")),  # use supported backends
    cls=clickext.ConfigOption,
    required=False,  # auto-detect
)
@click.option(
    "--gpus",
    type=click.IntRange(min=0),
    cls=clickext.ConfigOption,
    config_sections="vllm",
)
@click.pass_context
@clickext.display_params
def serve(
    ctx: click.Context,
    model_path: pathlib.Path,
    gpu_layers: int,
    num_threads: int | None,
    max_ctx_size: int,
    model_family,
    log_file: pathlib.Path | None,
    backend: str | None,
    chat_template: str | None,
    gpus: int | None,
) -> None:
    """Starts a local server"""

    warn_for_unsupported_backend_param(ctx)

    serve_backend(
        ctx,
        model_path,
        gpu_layers,
        num_threads,
        max_ctx_size,
        model_family,
        log_file,
        backend,
        chat_template,
        gpus,
    )


if __name__ == "__main__":
    serve()
